/* dga.c
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * This is the XFree86 DGA driver for PETAL.
 */

// Standard Unix includes
#include <stdio.h>

// The X-Windows includes (MIT-SHM and DGA)
#include <X11/X.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/xf86dga.h>

#include "../petal.h"
#include "dga.h"

Display *display;
Screen *screen;
Window window;
GC gc;

XEvent event;
char button;
int px, py;

char *address;
int  width;
int  bytes_per_line;
int  height;
int  bank_size;

int detect(void) {
  int major, minor, event_base, error_base;

  // Try to connect to the X Server
  display = XOpenDisplay(NULL);
  if( display == NULL ) return 1;

  if(!XF86DGAQueryVersion(display,&major,&minor)) {
    printf("DGA version query failed\n");
    return 1;
  }

  if(!XF86DGAQueryExtension(display,&event_base,&error_base)) {
    printf("DGA extension query failed\n");
    return 1;
  }

  /* Fail if the extension version in the server is too old */
  if(major < 1 || (major == 1 && minor < 0)) {
    printf("DGA extension is too old, get version 1.0 or higher\n");
    return 1;
  }
  else printf("Using DGA version %d.%d\n", major, minor);

  // Flag that we're using the MIT-SHM driver
  petal.magic = DGA;
  return 0;
}

int set_mode(char *window_name, int x, int y, int bpp) {
  XSetWindowAttributes xswa;
  XWindowAttributes    xwa;
  XPixmapFormatValues *pmf;
  int ram, mem_size, i, n;
  unsigned long mask;
  int width, height;

  XF86DGAGetViewPortSize(display, DefaultScreen(display), &width, &height);
  screen = DefaultScreenOfDisplay(display);
//visual = DefaultVisual(display, DefaultScreen(display));
  gc = XDefaultGC(display,DefaultScreen(display));

  // override redirect tells the window manger to just ignore us
  xswa.override_redirect = True;
  xswa.event_mask = ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
                    StructureNotifyMask | KeyPressMask | KeyReleaseMask;
  xswa.background_pixel = BlackPixelOfScreen(screen);
  mask = CWEventMask | CWOverrideRedirect | CWBackPixel;

  // Create the window
  //printf("Creating window\n");
  window = XCreateWindow(display, DefaultRootWindow(display), 0, 0,
	  	 	 width, height, 0, CopyFromParent, InputOutput,
		 	 CopyFromParent, mask, &xswa);

  // Setup some properties
  //XSetStandardProperties(display, window, window_name, window_name, None, NULL, 0, NULL);

  if(!window) {
    printf("Unable to open window.\n");
    return 1;
  }

  XMapWindow(display,window);
  XRaiseWindow(display,window);

  // shanghai the keyboard
  //printf("Grabbing mouse and keyboard\n");
  XGrabKeyboard(display, window, True, GrabModeAsync,
	 	          GrabModeAsync, CurrentTime);
  // shanghai the mouse
  XGrabPointer(display, window, True, ButtonPressMask | PointerMotionMask,
               GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

  // here goes nothing!
  XF86DGAGetVideo(display,DefaultScreen(display), &address, &bytes_per_line, &bank_size, &ram);
  XF86DGADirectVideo(display,DefaultScreen(display),
                     XF86DGADirectGraphics /*| XF86DGADirectMouse | XF86DGADirectKeyb*/);

  //printf("MemBase = 0x%x, Width = %i, BankSize = %i\n", address, bytes_per_line, bank_size);

  /*
  if(DoubleBuffer && ((PhysWidth*Height*2*Bpp) > (ram * 1024))) {
     fprintf(stderr,"Not enough memory to DoubleBuffer at this resolution.\n"); 
     fprintf(stderr,"\tTurning off DoubleBuffering.\n");
     DoubleBuffer = 0;
  }
  */

  if(bank_size < (ram * 1024)) {
    printf("Banked memory... Which I don't support yet... sorry!");
    // Banked memory, so set page?
    XF86DGASetVidPage(display,DefaultScreen(display),0);
    //memory_size = (byte_per_line*height*bpp);
    //if(BankSize < MemoryRequired) {
    //  BankSwitching = (MemoryRequired/BankSize);
    //  if((BankSwitching * BankSize) < (MemoryRequired))	BankSwitching++;
    //  printf("Window spans %i pages.\n",BankSwitching);
    //}
    //else printf("Window fits in one page,\n");
  }
  
  //printf("Setting viewport\n");
  XF86DGASetViewPort(display,DefaultScreen(display),0,0);

  // Setup some info about the mode we're in
  XGetWindowAttributes(display,window,&xwa);

  // Setup some info about the mode we're in
  petal.display.width = width;
  petal.display.height = height;

  // Now we have to search through the different pixmap formats available
  // to find the correct bits_per_pixel value for our colour depth.  Again,
  // remember, the colour depth is _NOT_ always equal to the bits/pixel.
  pmf = XListPixmapFormats(display, &n);
  for(i = 0; i < n; i++) {
    if(pmf[i].depth == xwa.depth) {
      petal.display.bpp = pmf[i].bits_per_pixel;
      break;
    }
  }
  XFree((char *)pmf);
  if(xwa.depth == 15) {
    petal.display.bytes_per_pixel = 2;
    petal.display.bpp = 15;
  }
  else petal.display.bytes_per_pixel = (petal.display.bpp >> 3);
  petal.display.bytes_per_line = bytes_per_line * petal.display.bytes_per_pixel;
  petal.display.size = petal.display.height * petal.display.bytes_per_line;

  return 0;
}

void deinit(void) {
  XF86DGADirectVideo(display, DefaultScreen(display), 0);
  XFreeGC(display, gc);
  XDestroyWindow(display, window);
  XUngrabPointer(display, CurrentTime);
  XUngrabKeyboard(display, CurrentTime);
  //XCloseDisplay(display);

  return;
}

// This function is used to get a pixel value at a certain location.  More
// specifically, it's used in the transparent blits to determine if a pixel is
// transparent or not.
/*
int get_color(char *text, int bytes) {
  int c;
  switch(bytes) {
    case 1:
      __asm__ (
        "movb (%%eax), %%bl"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x000000FF;
    case 2:
      __asm__ (
        "movw (%%eax), %%bx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x0000ffff;
    case 3:
      __asm__ (
        "movl (%%eax), %%ebx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x00FFFFFF;
    case 4:
      __asm__ (
        "movl (%%eax), %%ebx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c;
  }
  return 0;
}
*/

/* This blit function is VERY complicated, it seems.  Much more than I would've
 * hoped for, anyway.  I've tried to make it as fast as possible but you've got
 * to expect some slow down due the extreme versitility of PETAL's Image object.
 * PETAL allows images of any size, any location, with offsets, clipping, and
 * transparency to be blitting!
 */
void blit(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;
  int tx1, tx2, t, c;

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen.  We'll also use the
  // clipping technique if the image is transparent.  This
  // isn't necessary, but it's easier and takes less code.
  clipping  = img->clipping | img->transparent | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // Clip the image (found in petal.h)
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then the width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate starting offsets for the image and display (bottom left corner
    // so we can work from the bottom up)
    img_offset = (y1+ pixel_height) * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    // copy clipped image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      img_offset -= img->bytes_per_line;
      dsp_offset -= petal.display.bytes_per_line;

      // And display it (using either the transparent or full method)
      if(img->transparent) {
        // these will hold non-transparent spans to copy...
        tx1 = tx2 = x1 * img->bytes_per_pixel;
        t = img->trans_col;         // the transparent colour which will not be copied
        while(tx1 < width) {
          // Try and find the first non-transparent pixel
          c = get_color(&img->address[img_offset + tx1], img->bytes_per_pixel);
          if(c != t) {
            tx2 = tx1;  // start at the beginning of the non-transparent span
            // and try and find the end of it...
            while(get_color(&img->address[img_offset + tx2], img->bytes_per_pixel) != t &&
                  tx2 < width)
              tx2 += img->bytes_per_pixel;  // we haven't found the end yet so try the next pixel
            // Now draw the span
            memcpy((char *)(address + dsp_offset + tx1), &img->address[img_offset + tx1], tx2-tx1);
            tx1 = tx2;                     // search for the next span after this one, ofcourse
          }
          tx1 += img->bytes_per_pixel;    // goto the next pixel (haven't found start of span yet)
        }
      }
      else  // Not transparent, still clipped...
        memcpy((char *)(address + dsp_offset), &img->address[img_offset], width);
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    memcpy(address, img->address, img->size);
  }
  return;
}

/* Well, this function should look a bit familar.  It's almost exactly
 * the same as blit, except it works in reverse.
 */
void get_image(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen
  clipping  = img->clipping | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // Perform some advanced clipping
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then a width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate starting offsets for the image and display (bottom left corner
    // so we can work from the bottom up)
    img_offset = (y1 +pixel_height) * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    // copy clipped image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      img_offset -= img->bytes_per_line;
      dsp_offset -= petal.display.bytes_per_line;
      memcpy(&img->address[img_offset], (char *)(address + dsp_offset), width);
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    memcpy(img->address, address, img->size);
  }
  return;
}

char get_pointer(int *x, int *y) {
  while( XCheckTypedEvent(display, MotionNotify, &event) ) {
    px = event.xmotion.x_root;
    py = event.xmotion.y_root;
  }

  while(	XCheckTypedEvent(display, ButtonPress, &event) )
    button |= event.xbutton.button;

  while(	XCheckTypedEvent(display, ButtonRelease, &event) )
    button &= !event.xbutton.button;

  // The following are actually possible under DGA (panned screen)
  if(px < 0) px = 0;
  if(py < 0) py = 0;
  if(px > petal.display.width) px = petal.display.width;
  if(py > petal.display.height) py = petal.display.height;

  *x = px; *y = py;
  return button;
}

char get_key(void) {
  char key = 0;

  while( XCheckTypedEvent(display, KeyPress, &event) )
    key = XKeycodeToKeysym(display, event.xkey.keycode, 0);

  return key;
}

void set_pal(rgb *pal) {
  XColor temp[256];
  Colormap cmap;
  int i;

  for(i = 0; i <= 255; i++) {
    temp[i].pixel = i;
    temp[i].flags = DoRed | DoGreen | DoBlue;
    temp[i].red   = (int)pal[i].r * 256;
    temp[i].green = (int)pal[i].g * 256;
    temp[i].blue  = (int)pal[i].b * 256;
  }
  cmap=XCreateColormap(display, RootWindowOfScreen(screen),
                       DefaultVisualOfScreen(screen), AllocAll);
  XStoreColors(display, cmap, temp, 255);
  XF86DGAInstallColormap(display, DefaultScreen(display), cmap);
}
