#ifdef __DJGPP__

#include <dos.h>
#include <go32.h>
#include <dpmi.h>
#include <stdlib.h>
#include <string.h>

#define MIN(a,b)   ( ((a) < (b)) ? (a) : (b) )
#define MAX(a,b)   ( ((a) > (b)) ? (a) : (b) )
#define MID(a,b,c) MAX( (a), MIN( (b) , (c) ))

volatile int mouse_x = 0;           /* mouse x pos */
volatile int mouse_y = 0;           /* mouse y pos */
volatile int mouse_b = 0;           /* mouse button state */

static int mouse_mx = 0;            /* internal position, in mickeys */
static int mouse_my = 0;

static int mouse_sx = 2;            /* mickey -> pixel scaling factor */
static int mouse_sy = 2;

int mouse_minx = 0;                 /* mouse range */
int mouse_miny = 0;
int mouse_maxx = 320;
int mouse_maxy = 200;

static int mouse_installed = 0;

#define MICKEY_TO_COORD_X(n)        ((n) / mouse_sx)
#define MICKEY_TO_COORD_Y(n)        ((n) / mouse_sy)

#define COORD_TO_MICKEY_X(n)        ((n) * mouse_sx)
#define COORD_TO_MICKEY_Y(n)        ((n) * mouse_sy)

static _go32_dpmi_seginfo mouse_seginfo;
static _go32_dpmi_registers mouse_regs;

// My mouse interupt, borrowed from Allegro, with quite a bit of
// modification.
static void mouseint(_go32_dpmi_registers *r) {
  int mx, my;
  __dpmi_regs regs;

  mouse_b = r->x.bx;
  mx = (signed short)r->x.si;
  my = (signed short)r->x.di;

  mouse_x = MICKEY_TO_COORD_X(mouse_mx + mx);
  mouse_y = MICKEY_TO_COORD_Y(mouse_my + my);

  if((mouse_x < mouse_minx) || (mouse_x > mouse_maxx) ||
     (mouse_y < mouse_miny) || (mouse_y > mouse_maxy)) {
    mouse_x = MID(mouse_minx, mouse_x, mouse_maxx);
	 mouse_y = MID(mouse_miny, mouse_y, mouse_maxy);

	 mouse_mx = COORD_TO_MICKEY_X(mouse_x);
	 mouse_my = COORD_TO_MICKEY_Y(mouse_y);

    // clear the mickeys
    regs.x.ax = 11;
    __dpmi_int(0x33, &regs);
  }
}


// Install the mouse handler
void install_mouse(void) {
  __dpmi_regs r;
  int num_buttons;

  if (mouse_installed) return;

  r.x.ax = 0;
  __dpmi_int(0x33, &r);         /* initialise mouse driver */

  if(r.x.ax == 0) return;

  num_buttons = r.x.bx;
  if(num_buttons == 0xFFFF) num_buttons = 2;

  mouse_seginfo.pm_offset = (int)mouseint;
  mouse_seginfo.pm_selector = _my_cs();
  _go32_dpmi_allocate_real_mode_callback_retf(&mouse_seginfo, &mouse_regs);

  r.x.ax = 0x0C;  // mouse interrupt
  r.x.cx = 0x7F;
  r.x.dx = mouse_seginfo.rm_offset;
  r.x.es = mouse_seginfo.rm_segment;
  __dpmi_int(0x33, &r);            /* install callback */

  mouse_installed = 1;

  return;
}

// remove the mouse handler
void remove_mouse(void) {
  __dpmi_regs r;

  if (!mouse_installed) return;

  r.x.ax = 0x0C;
  r.x.cx = 0;
  r.x.dx = 0;
  r.x.es = 0;
  __dpmi_int(0x33, &r);         /* install NULL callback */

  mouse_x = mouse_y = mouse_b = 0;
  _go32_dpmi_free_real_mode_callback(&mouse_seginfo);
  mouse_installed = 0;
}

#endif
