/* colour.cc
 * 1998/02/13
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * I've separated a couple functions here because they're
 * also used by the drivers.  This file my be linked to
 * any PETAL programs, and any PETAL drivers.  Unfortunately,
 * you'll have to keep two copies of it;  One as C++ and
 * one as C.  The reason?  The X11 drivers must be C
 * files.
 */

#include "../petal.h"

int clip(Image *one, Image *two, signed int *x, signed int *y, int *x1, int *y1, int *x2, int *y2) {
  /* Set the initial clipping values to the smallest of the two windows */
  *x1 = *y1 = 0;
  *x2 = MIN(one->width,  two->width);
  *y2 = MIN(one->height, two->height);

  /* Clip to the first window, if we're told to */
  if(one->clipping) {
    if(one->cx1 > *x1) *x1 = one->cx1;
    if(one->cy1 > *y1) *y1 = one->cy1;
    if(one->cx2 < *x2) *x2 = one->cx2;
    if(one->cy2 < *y2) *y2 = one->cy2;
  }

  /* Clip to the second window, if we're told to */
  if(two->clipping) {
    if(two->cx1 > *x1) *x1 = two->cx1;
    if(two->cy1 > *y1) *y1 = two->cy1;
    if(two->cx2 < *x2) *x2 = two->cx2;
    if(two->cy2 < *y2) *y2 = two->cy2;
  }

  /* After clipping, we adjust our clip window by the image's offsets to
   * locate it in the correct position.
   */
  OffsetAdjust(one, *x1, *y1);
  OffsetAdjust(one, *x2, *y2);

  /* Also remember to adjust the x and y locations by the display offsets */
  *x += two->x_off;
  *y += two->y_off;
  
  /* If the display offsets end up being negative we'll shift our
   * object up and to the left to accomodate.
   */
  if(*x < 0) { *x1 -= *x; *x = 0; }
  if(*y < 0) { *y1 -= *y; *y = 0; }

  /* After all that, check if we're even in bounds!  If not, quit. */
  if(*x  > two->width || *y  > two->height) return 0;   /* off screen bottom right  */
  if(*x1 > one->width || *y1 > one->height) return 0;   /* off screen from top-left */
  /*if(*x1 < 0 || *y1 < 0) return False;*/              /* will never happen?       */

  /* Find an acceptable width */
  if((*x2 - *x1) + *x > two->width)  *x2 = two->width - *x;

  /* Find an acceptable height */
  if((*y2 - *y1) + *y > two->height) *y2 = two->height - *y;

  return 1;
}

/* Used by drivers as a getpixel-like function.  More-or-less a hack
 * and will be fixed in the next version of PETAL (1.2)
 */
int get_color(char *text, int bytes) {
  int c;
  switch(bytes) {
    case 1:
      __asm__ (
        "movb (%%eax), %%bl"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x000000FF;
    case 2:
      __asm__ (
        "movw (%%eax), %%bx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x0000ffff;
    case 3:
      __asm__ (
        "movl (%%eax), %%ebx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x00FFFFFF;
    case 4:
      __asm__ (
        "movl (%%eax), %%ebx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c;
  }
  return 0;
}
