/* lfb.c
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * This is the Vesa 2.0 linear frame buffer
 * driver for PETAL.
 */

#include "../petal.h"
#include "lfb.h"
#include "vesa.h"

#ifdef __DJGPP__

// Standard DJGPP includes
#include <dos.h>
#include <dpmi.h>
#include <go32.h>
#include <stdio.h>
#include <conio.h>
#include <assert.h>
#include <sys/farptr.h>
#include <sys/nearptr.h>

extern volatile int mouse_x;
extern volatile int mouse_y;
extern volatile int mouse_b;
extern volatile int mouse_pos;

extern int mouse_maxx;
extern int mouse_maxy;

void remove_mouse(void);
void install_mouse(void);

Driver lfb_driver = {
 LFB, 1, 2, 3, 4, 5, 6, NULL, 
 // Image display starts here...
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, NULL,
 0,
 // ... Image display ends here
 NULL, NULL, NULL,
 lfb_detect,
 lfb_set_mode,
 lfb_deinit,
 lfb_blit,
 lfb_get_image,
 lfb_get_pointer,
 lfb_get_key,
 lfb_set_pal
};

unsigned int linear_address;

int lfb_detect(void) {
  // I refine the below because, just in case the detect fails, I want
  // the program to be able to run in the previous mode (if it was in
  // one) unharmed, and that means keeping the structures alive.  It's
  // okay if we overwrite the card structure (whos going to change their
  // video card in the middle of a program?) and much more convenient.
  VesaModeInfo modeinfo;

  // turn vesa ON for everex cards! :)
  regs.x.ax = 0x4FFF;
  regs.h.dl = 0x1;
  __dpmi_int(0x10,&regs);

  // get some vesa information
  assert(sizeof(card) < _go32_info_block.size_of_transfer_buffer);
  strncpy(card.VESASignature, "VBE2",4);
  regs.x.ax = 0x4F00;
  regs.x.di = __tb & 0x0F;
  regs.x.es = (__tb >> 4) & 0xFFFF;
  dosmemput(&card,sizeof(card),__tb);
  __dpmi_int(0x10, &regs);
  dosmemget(__tb, sizeof(card), &card);

  if(strncmp(card.VESASignature,"VESA",4) == 0)   // it's a vesa card
    if(card.VESAVersion >= 0x200) {               // it's vesa 2.0
      // get some info on an arbitrary mode and check if an LFB is found
      // (chances are if we can find an LFB is one mode, it'll be available
      //  in others.  This is all we can do for detection).
      regs.x.ax = 0x4F01;
      regs.x.di = __tb & 0x0F;             // the offset of the transfer buffer
      regs.x.es = (__tb >> 4) & 0xFFFF;    // the segment of the transfer buffer
      regs.x.cx = 0x101;                   // arbitrary mode
      __dpmi_int(0x10, &regs);
      dosmemget(__tb, sizeof(modeinfo), &modeinfo);

      if(modeinfo.ModeAttributes & 0x8) return 0;
      else return 1;
    }
  return 1;
}

int lfb_set_mode(char *name, int x, int y, int bpp) {
  // get a protected mode address from the video mode pointer (which is a
  // real mode far pointer).
  long mode_ptr = ((card.VideoModePtr & 0xFFFF0000) >> 12) +
                   (card.VideoModePtr & 0xFFFF);
  unsigned short mode_num;
  __dpmi_meminfo mem;

  __djgpp_nearptr_enable();   // enable near pointers from now on

  if(x == 320 && y == 200 && bpp == 8) {
    // we need some special code for VGA mode 13h because it's not a Vesa
    // mode and therefore will not be recognised as a valid mode by the Vesa
    // bios (and we want to support it).
    regs.x.ax = 0x0013;
    __dpmi_int(0x10,&regs);
    petal.display.width  = mouse_maxx = x;
    petal.display.height = mouse_maxy = y;
    petal.display.bpp = bpp;
    petal.display.bytes_per_pixel = bytes_per_pixel = 1;
    petal.display.bytes_per_line = bytes_per_line = 320;
    petal.display.size = 64000;
    linear_address = 0xA0000;          // linear address of vider memory
    install_mouse();
    return 0;
  }

  // the video mode list is terminated by 0xFFFF.  Loop until we reach it.
  while((mode_num = _farpeekw(_dos_ds, mode_ptr)) != 0xFFFF) {
    mode_ptr += 2;                     // go the next mode (+2 bytes = +short)
    get_mode_info(mode_num, &mode);    // get some info about this mode
    // check to see if it's the mode we want
    if((int)mode.XResolution  == x   &&
       (int)mode.YResolution  == y   &&
       (int)mode.BitsPerPixel == bpp &&
       (mode.ModeAttributes & 0x1) != 0  &&  // mode is available?
       (mode.ModeAttributes & 0x8) != 0) {   // LFB available?
         // get the LFB and convert it to a linear address
         // Be aware: overwrite and you WILL get a page fault.  Using banked
         // memory, you rarely get a page fault... even in protected mode.  I
         // got tons of page faults when writting this, so I add 5 for good
         // luck here :)
         mem.size = (unsigned long)((int)mode.YResolution *
                                    (int)mode.BytesPerScanLine) + 5;
         mem.address = mode.PhysBasePtr;
         __dpmi_physical_address_mapping(&mem);
         linear_address = mem.address;
         //address = (unsigned char *)linear_address;
         // set the mode
         regs.x.ax = 0x4F02;                 // set vesa mode
         regs.x.bx = (mode_num | 0x4000);    // set bit 15 for LFB access
         __dpmi_int(0x10,&regs);
         petal.display.width  = mouse_maxx = x;
         petal.display.height = mouse_maxy = y;
         petal.display.bpp    = bpp;
         if(bpp == 15) bytes_per_pixel = 2;  // last bit is ignored
         else bytes_per_pixel = (bpp/8);     // I hope that'll work for all 24-bpp modes
         // Believe it or not, BytesPerScanLine is NOT always equal to the
         // bytes/pixel * XResolution!  That one was a headache and a half!
         bytes_per_line = mode.BytesPerScanLine;
         petal.display.bytes_per_pixel = bytes_per_pixel;
         petal.display.bytes_per_line  = bytes_per_line;
         petal.display.size  = mode.YResolution * mode.BytesPerScanLine;
         install_mouse();
         return 0;                     // found specified mode
      }
  }
  return 1;                            // specified mode not found
}

void lfb_deinit(void) {
  regs.x.ax = 0x3;
  __dpmi_int(0x10, &regs);
  __djgpp_nearptr_disable();   // disable near pointers before leaving
  remove_mouse();
  return;
}

/* This blit function is VERY complicated, it seems.  Much more than I would've
 * hoped for, anyway.  I've tried to make it as fast as possible but you've got
 * to expect some slow down due the extreme versitility of PETAL's Image object.
 * PETAL allows images of any size, any location, with offsets, clipping, and
 * transparency to be blitting!
 */
void lfb_blit(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;
  int tx1, tx2, t, c;

  char *address = (char*)(linear_address+__djgpp_conventional_base);

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen.  We'll also use the
  // clipping technique if the image is transparent.  This
  // isn't necessary, but it's easier and takes less code.
  clipping  = img->clipping | img->transparent | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // Clip the image (found in petal.h)
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then the width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate starting offsets for the image and display (bottom left corner
    // so we can work from the bottom up)
    img_offset = (y1+ pixel_height) * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    // copy clipped image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      img_offset -= img->bytes_per_line;
      dsp_offset -= petal.display.bytes_per_line;

      // And display it (using either the transparent or full method)
      if(img->transparent) {
        // these will hold non-transparent spans to copy...
        tx1 = tx2 = x1 * img->bytes_per_pixel;
        t = img->trans_col;         // the transparent colour which will not be copied
        while(tx1 < width) {
          // Try and find the first non-transparent pixel
          c = get_color(&img->address[img_offset + tx1], img->bytes_per_pixel);
          if(c != t) {
            tx2 = tx1;  // start at the beginning of the non-transparent span
            // and try and find the end of it...
            while(get_color(&img->address[img_offset + tx2], img->bytes_per_pixel) != t &&
                  tx2 < width)
              tx2 += img->bytes_per_pixel;  // we haven't found the end yet so try the next pixel
            // Now draw the span
            memcpy((char *)(address + dsp_offset + tx1), &img->address[img_offset + tx1], tx2-tx1);
            tx1 = tx2;                     // search for the next span after this one, ofcourse
          }
          tx1 += img->bytes_per_pixel;    // goto the next pixel (haven't found start of span yet)
        }
      }
      else  // Not transparent, still clipped...
        memcpy((char *)(address + dsp_offset), &img->address[img_offset], width);
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    memcpy(address, img->address, img->size);
  }
  return;
}

/* Well, this function should look a bit familar.  It's almost exactly
 * the same as blit, except it works in reverse.
 */
void lfb_get_image(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;

  char *address = (char*)(linear_address+__djgpp_conventional_base);

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen
  clipping  = img->clipping | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // Perform some advanced clipping
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then a width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate starting offsets for the image and display (bottom left corner
    // so we can work from the bottom up)
    img_offset = (y1 +pixel_height) * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    // copy clipped image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      img_offset -= img->bytes_per_line;
      dsp_offset -= petal.display.bytes_per_line;
      memcpy(&img->address[img_offset], (char *)(address + dsp_offset), width);
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    memcpy(img->address, address, img->size);
  }
  return;
}

char lfb_get_pointer(int *x, int *y) {
  *x = mouse_x;
  *y = mouse_y;
  return mouse_b;
}

char lfb_get_key(void) {
  if(kbhit()) return getch();
  else return 0;
}

void lfb_set_pal(rgb *pal) {
  int i;

  outportb(0x3c8,0);
  for(i = 0; i < 256; i++) {
    outportb(0x3c9, (pal[i].r >> 2));
    outportb(0x3c9, (pal[i].g >> 2));
    outportb(0x3c9, (pal[i].b >> 2));
  }
  return;
}

#else

Driver lfb_driver;

#endif
