/* pmode.c
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * This is the Vesa 2.0 protected mode banking
 * driver for PETAL.
 */

#include "../petal.h"
#include "pmode.h"
#include "vesa.h"

#ifdef __DJGPP__

// Standard DJGPP includes
#include <dos.h>
#include <dpmi.h>
#include <go32.h>
#include <stdio.h>
#include <conio.h>
#include <assert.h>
#include <memory.h>
#include <sys/farptr.h>
#include <sys/nearptr.h>

extern volatile int mouse_x;
extern volatile int mouse_y;
extern volatile int mouse_b;
extern volatile int mouse_pos;

extern int mouse_maxx;
extern int mouse_maxy;

void remove_mouse(void);
void install_mouse(void);

inline void pm_set_bank(int bank);

Driver pmode_driver = {
 PMODE, 1, 2, 3, 4, 5, 6, NULL,
 // Image display starts here...
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, NULL,
 0,
 // ... Image display ends here
 NULL, NULL, NULL,
 pmode_detect,
 pmode_set_mode,
 pmode_deinit,
 pmode_blit,
 pmode_get_image,
 pmode_get_pointer,
 pmode_get_key,
 pmode_set_pal
};

VesaProtected *pm_info;
void (*pm_bank)();
void (*pm_scroll)();
void (*pm_palette)();

int pmode_detect(void) {
  // I refine the below because, just in case the detect fails, I want
  // the program to be able to run in the previous mode (if it was in
  // one) unharmed, and that means keeping the structures alive.  It's
  // okay if we overwrite the card structure (whos going to change their
  // video card in the middle of a program?) and much more convenient.
  VesaModeInfo modeinfo;

  // turn vesa ON for everex cards! :)
  regs.x.ax = 0x4FFF;
  regs.h.dl = 0x1;
  __dpmi_int(0x10,&regs);

  // get some vesa information
  assert(sizeof(card) < _go32_info_block.size_of_transfer_buffer);
  strncpy(card.VESASignature, "VBE2",4);
  regs.x.ax = 0x4F00;
  regs.x.di = __tb & 0x0F;
  regs.x.es = (__tb >> 4) & 0xFFFF;
  dosmemput(&card,sizeof(card),__tb);
  __dpmi_int(0x10, &regs);
  dosmemget(__tb, sizeof(card), &card);

  if(strncmp(card.VESASignature,"VESA",4) == 0)   // it's a vesa card
    if(card.VESAVersion >= 0x200) {               // it's vesa 2.0
      regs.x.ax = 0x4F0A;                         // get pmode interface...
      regs.x.bx = 0;
      __dpmi_int(0x10, &regs);
      if(regs.h.ah) return 1;                     // no interface found

      pm_info = (VesaProtected *)malloc(regs.x.cx); // allocate enough room for it
      _go32_dpmi_lock_data(pm_info, regs.x.cx);     // lock it
      dosmemget((regs.x.es*16)+regs.x.di, regs.x.cx, pm_info); // and retrieve it

      // fill in our protected mode functions with some code
      pm_bank     = (void (*)()) ((char *)pm_info + pm_info->setWindow);
      pm_scroll   = (void (*)()) ((char *)pm_info + pm_info->setDisplayStart);
      pm_palette  = (void (*)()) ((char *)pm_info + pm_info->setPalette);

      return 0;                                   // yep... pmode interface found
    }
  return 1;
}

int pmode_set_mode(char *name, int x, int y, int bpp) {
  // get a protected mode address from the video mode pointer (which is a
  // real mode far pointer).
  long mode_ptr = ((card.VideoModePtr & 0xFFFF0000) >> 12) +
                   (card.VideoModePtr & 0xFFFF);
  unsigned short mode_num;

  __djgpp_nearptr_enable();   // enable near pointers from now on

  if(x == 320 && y == 200 && bpp == 8) {
    // we need some special code for VGA mode 13h because it's not a Vesa
    // mode and therefore will not be recognised as a valid mode by the Vesa
    // bios (and we want to support it).
    regs.x.ax = 0x0013;
    __dpmi_int(0x10,&regs);
    petal.display.width  = mouse_maxx = x;
    petal.display.height = mouse_maxy = y;
    petal.display.bpp = bpp;
    petal.display.bytes_per_pixel = bytes_per_pixel = 1;
    petal.display.bytes_per_line = bytes_per_line = 320;
    petal.display.size = 64000;
    install_mouse();
    return 0;
  }

  // the video mode list is terminated by 0xFFFF.  Loop until we reach it.
  while((mode_num = _farpeekw(_dos_ds, mode_ptr)) != 0xFFFF) {
    mode_ptr += 2;                     // go the next mode (+2 bytes = +short)
    get_mode_info(mode_num, &mode);    // get some info about this mode
    // check to see if it's the mode we want
    if((int)mode.XResolution  == x   &&
       (int)mode.YResolution  == y   &&
       (int)mode.BitsPerPixel == bpp &&
       (mode.ModeAttributes & 0x1) != 0) {  // mode is available?
         // if it is, then let's use it!
         regs.x.ax = 0x4F02;
         regs.x.bx = mode_num;
         __dpmi_int(0x10,&regs);
         // setup some important info
         petal.display.width  = mouse_maxx = x;
         petal.display.height = mouse_maxy = y;
         petal.display.bpp    = bpp;
         if(bpp == 15) bytes_per_pixel = 2;  // last bit is ignored
         else bytes_per_pixel = (bpp/8);     // I hope that'll work for all 24-bpp modes
         // Believe it or not, BytesPerScanLine is NOT always equal to the
         // bytes/pixel * XResolution!  That one was a headache and a half!
         bytes_per_line = mode.BytesPerScanLine;
         petal.display.bytes_per_pixel = bytes_per_pixel;
         petal.display.bytes_per_line  = bytes_per_line;
         petal.display.size  = mode.YResolution * mode.BytesPerScanLine;
         install_mouse();
         return 0;                     // found specified mode
      }
  }
  return 1;                            // specified mode not found
}

void pmode_deinit(void) {
  regs.x.ax = 0x3;
  __dpmi_int(0x10, &regs);
  __djgpp_nearptr_disable();   // disable near pointers before leaving
  remove_mouse();
  return;
}

/* This blit function is VERY complicated, it seems.  Much more than I would've
 * hoped for, anyway.  I've tried to make it as fast as possible but you've got
 * to expect some slow down due the extreme versitility of PETAL's Image object.
 * PETAL allows images of any size, any location, with offsets, clipping, and
 * transparency to be blitting!
 */
void pmode_blit(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;
  int tx1, tx2, t, c;
  int bank_offset, bank;
  int bwidth1, bwidth2;

  char *address = (char*)(0xA0000+__djgpp_conventional_base);

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen.  We'll also use the
  // clipping technique if the image is transparent.  This
  // isn't necessary, but it's easier and takes less code.
  clipping  = img->clipping | img->transparent | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // Clip the image (found in petal.h)
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then the width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate starting offsets for the image and display
    img_offset = y1 * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = y  * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    pm_set_bank((int)(dsp_offset >> 16));
    bank_offset = (dsp_offset & 0xFFFF);

    // copy clipped image to screen
    for(i = 0; i < pixel_height; i++) {
      // And display it (using either the transparent or full method)
      if(img->transparent) {
        // these will hold non-transparent spans to copy...
        tx1 = tx2 = x1 * img->bytes_per_pixel;
        t = img->trans_col;         // the transparent colour which will not be copied
        while(tx1 < width) {
          // Try and find the first non-transparent pixel
          c = get_color(&img->address[img_offset + tx1], img->bytes_per_pixel);
          if(c != t) {
            tx2 = tx1;  // start at the beginning of the non-transparent span
            // and try and find the end of it...
            while(get_color(&img->address[img_offset + tx2], img->bytes_per_pixel) != t &&
                  tx2 < width) tx2 += img->bytes_per_pixel;  // we haven't found the end yet so try the next pixel

            // Now draw the span
            bank_offset = ((dsp_offset + tx1) & 0xFFFF);
            pm_set_bank((dsp_offset + tx1) >> 16);
            if(bank_offset + (tx2-tx1) > 65536) {
              bwidth1 = 65536 - bank_offset;
              bwidth2 = (bank_offset + (tx2-tx1)) - 65536;
              // Draw what we can...
              memcpy((char *)(address + bank_offset), &img->address[img_offset+tx1], bwidth1);
              // Advance to the next bank...
              pm_set_bank(curbank + 1);
              bank_offset = 0;
              // And draw the rest...
              memcpy((char *)(address), &img->address[(img_offset+bwidth1+tx1)], bwidth2);
              //memcpy((char *)(address + bank_offset), &img->address[img_offset + tx1], tx2-tx1);
              //pm_set_bank(curbank + 1);
              //bwidth = ((tx2 - tx1) | 0xFFFF) - (tx2 - tx1);
              //bank_offset = 0;
              //memcpy((char *)(address + bank_offset), &img->address[img_offset + tx1], bwidth);
            }
            else
              memcpy((char *)(address + bank_offset), &img->address[img_offset + tx1], tx2-tx1);
            tx1 = tx2;                     // search for the next span after this one, ofcourse
          }
          tx1 += img->bytes_per_pixel;    // goto the next pixel (haven't found start of span yet)
        }
      }
      else { // Not transparent, still clipped...
        bank_offset = dsp_offset & 0xFFFF;
        pm_set_bank(dsp_offset >> 16);
        if(bank_offset + width > 65536) {
          bwidth1 = 65536 - bank_offset;
          bwidth2 = (bank_offset + width) - 65536;
          // Draw what we can...
          memcpy((char *)(address + bank_offset), &img->address[img_offset], bwidth1);
          // Advance to the next bank...
          pm_set_bank(curbank + 1);
          bank_offset = 0;
          // And draw the rest...
          memcpy((char *)(address), &img->address[(img_offset+bwidth1)], bwidth2);
        }
        else
          memcpy((char *)(address + bank_offset), &img->address[img_offset], width);
      }
      // advance to the next line
      img_offset += img->bytes_per_line;
      dsp_offset += petal.display.bytes_per_line;
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    for(bank = 0; bank < petal.display.size; bank += 65536) {
      pm_set_bank((int)(bank >> 16));
      memcpy(address, &img->address[bank], 65536);
    }
  }
  return;
}

/* Well, this function should look a bit familar.  It's almost exactly
 * the same as blit, except it works in reverse.
 */
void pmode_get_image(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;
  int bank_offset, bank;
  int bwidth1, bwidth2;

  char *address = (char*)(0xA0000+__djgpp_conventional_base);

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen
  clipping  = img->clipping | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // Perform some advanced clipping
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then a width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate starting offsets for the image and display (bottom left corner
    // so we can work from the bottom up)
    img_offset = y1 * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = y  * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    pm_set_bank((int)(dsp_offset >> 16));
    bank_offset = (dsp_offset & 0xFFFF);

    // copy clipped image to screen (work from bottom to top for speed)
    for(i = 0; i < pixel_height; i++) {
      //bank_offset += width;
      bank_offset = dsp_offset & 0xFFFF;
      pm_set_bank(dsp_offset >> 16);

      if(bank_offset + width > 65536) {
        bwidth1 = 65536 - bank_offset;
        bwidth2 = (bank_offset + width) - 65536;
        // Draw what we can...
        memcpy(&img->address[img_offset],(char *)(address + bank_offset), bwidth1);
        // Advance to the next bank...
        pm_set_bank(curbank + 1);
        bank_offset = 0;
        // And draw the rest...
        memcpy(&img->address[(img_offset+bwidth1)], (char *)(address), bwidth2);
      }
      else
        memcpy(&img->address[img_offset], (char *)(address + bank_offset), width);

      img_offset += img->bytes_per_line;
      dsp_offset += petal.display.bytes_per_line;
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    for(bank = 0; bank < petal.display.size; bank += 65536) {
      pm_set_bank((int)(bank >> 16));
      memcpy(&img->address[bank], address, 65536);
    }
  }
  return;
}

char pmode_get_pointer(int *x, int *y) {
  *x = mouse_x;
  *y = mouse_y;
  return mouse_b;
}

char pmode_get_key(void) {
  if(kbhit()) return getch();
  else return 0;
}

inline void pm_set_bank(int bank_num) {
  if(bank_num != curbank) {           // Bank is already active
    curbank = bank_num;                // Save current bank number
    bank_num <<= mode.WinGranularity;  // Adjust to window granularity

    // that's AT&T for "mov dx, bank_num" "mov bx, 0" :)
    __asm__ volatile("
      xor %%ebx, %%ebx;
      call %%eax;"
      :
      : "d" (bank_num), "a" (pm_bank)
      : "bx"
    );
    //pm_bank();

    // this is AT&T for "mov dx, bank_num" "mov bx, 1" :)
    __asm__ volatile("
      mov $1, %%ebx;
      call %%eax;"
      :
      : "d" (bank_num), "a" (pm_bank)
      : "bx"
    );
    //__asm__ volatile("call %%eax" : : "d" (bank_num), "b" (1), "a" (pm_bank));
    //pm_bank();
  }
  return;
}

void pmode_set_pal(rgb *pal) {
  int i;

  outportb(0x3c8,0);
  for(i = 0; i < 256; i++) {
    outportb(0x3c9, (pal[i].r >> 2));
    outportb(0x3c9, (pal[i].g >> 2));
    outportb(0x3c9, (pal[i].b >> 2));
  }
  return;
}

#else

Driver pmode_driver;

#endif
