/* xshm.c
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * This is the MIT-SHM driver for PETAL.
 */

// Standard Unix includes
#include <stdio.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

// The X-Windows includes (MIT-SHM and DGA)
#include <X11/X.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/xf86dga.h>

#include "../petal.h"
#include "xshm.h"

// X11 variables
int drawable;                      // tells if is possible to draw
int CompletionType;
Display *display;
Screen *screen;
Window window;
Cursor cursor;
Pixmap shape, mask;
XColor fg, bg;  // fake colours for our fake cursor ;)
XColor temp[256];
Colormap cmap;
GC gc;
XImage *image;
XShmSegmentInfo shminfo;

XEvent event;
char button;
int px, py;

static char cursor_bits[] = { 0x00 };
static char mask_bits[] = { 0x00 };

int detect(void) {
  int major, minor, ignore, i;
  Bool pixmaps;

  // Try to connect to the X Server
  display = XOpenDisplay(NULL);
  if( display == NULL ) return 1;

  // Check for the XShm extension
  if( XQueryExtension(display, "MIT-SHM", &ignore, &ignore, &ignore) ) {
    if(XShmQueryVersion(display, &major, &minor, &pixmaps) == True) {
      printf("XShm extention version %d.%d %s shared pixmaps\n",
             major, minor, (pixmaps==True) ? "with" : "without");
    }
    else return 1;
  }

  // Flag that we're using the MIT-SHM driver
  petal.magic = XSHM;
  return 0;
}

int set_mode(char *window_name, int x, int y, int bpp) {
  XSetWindowAttributes xswa;

  // Get an instance of screen
  screen = DefaultScreenOfDisplay(display);

  // I want to be notified about the following events:
  xswa.event_mask = ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
                    StructureNotifyMask | KeyPressMask | KeyReleaseMask |
                    PointerMotionMask;
  xswa.background_pixel = BlackPixelOfScreen(screen);

  // Now, create my window:
  window = XCreateWindow(display, RootWindowOfScreen(screen), 0, 0, x, y, 0,
                         DefaultDepthOfScreen(screen), InputOutput, DefaultVisualOfScreen(screen),
                         CWEventMask | CWBackPixel, &xswa);

  // Setup some properties
  XSetStandardProperties(display, window, window_name, window_name, None, NULL, 0, NULL);

  // Create a graphics context
  gc = XCreateGC(display, window, 0, 0);

  // Map my window:
  XMapWindow(display, window);

  //XGrabKeyboard(display, window, True, GrabModeAsync,
  //	          GrabModeAsync, CurrentTime);

  //XGrabPointer(display, window, True, ButtonPressMask | PointerMotionMask,
  //             GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

  // detect() has already assured us that MIT-SHM exists,
  // so allocate an image and check for errors
  image = XShmCreateImage(display, DefaultVisualOfScreen(screen),
                          DefaultDepthOfScreen(screen),
                          ZPixmap, NULL, &shminfo,
                          x, y );
  if(image == NULL) return 1;

  // Get the shared memory and check for errors
  shminfo.shmid = shmget(IPC_PRIVATE, image->bytes_per_line*image->height,
                         IPC_CREAT | 0777 );
  if(shminfo.shmid < 0) return 1;

  // do something and check for errors :)
  shminfo.shmaddr = image->data = (char *)shmat(shminfo.shmid, 0, 0);
  if(shminfo.shmaddr == (char *) -1) return 1;

  // set as read only, and attach the display
  shminfo.readOnly = False;
  XShmAttach(display, &shminfo);

  // This will create an invisible cursor which we need to do for petal
  // because petal has it's own cursor control!
  shape = XCreateBitmapFromData(display, window, cursor_bits, 1, 1 ); // a 1x1 empty shape
  mask = XCreateBitmapFromData(display, window, mask_bits, 1, 1 );    // a 1x1 empty mask
  cursor = XCreatePixmapCursor(display, shape, mask, &fg, &bg, 0, 0 );
  XDefineCursor(display, window, cursor);

  // Used to tell if X11 is finished drawing
  CompletionType = XShmGetEventBase(display) + ShmCompletion;

  // Setup some info about the mode we're in.  It's really nice that the
  // XImage structure has a bits_per_pixel variable.  In the case of DGA
  // I have to go through a somewhat lengthy process to get to the
  // bits_per_pixel variable.  In any event, it should be noted that the
  // colour depth is _NOT_ always equal to the bits_per_pixel variable.
  // For example, my 24bpp screen has 32 bits/pixel.
  petal.display.width = x;
  petal.display.height = y;
  petal.display.bpp = image->bits_per_pixel;
  if(image->depth == 15) {
    petal.display.bytes_per_pixel = 2;
    petal.display.bpp = 15;
  }
  else petal.display.bytes_per_pixel = (petal.display.bpp >> 3);
  petal.display.bytes_per_line = image->bytes_per_line;
  petal.display.size = petal.display.height * petal.display.bytes_per_line;

  XFlush(display);
  return 0;                            // everything worked
}

void deinit(void) {
  XShmDetach(display, &shminfo);
  XDestroyImage(image);
  shmdt(shminfo.shmaddr);
  XFreeGC(display, gc);
  XDestroyWindow(display, window);
  //XUngrabPointer(display, CurrentTime);
  //XUngrabKeyboard(display, CurrentTime);
  //XCloseDisplay(display);
}

// This function is used to get a pixel value at a certain location.  More
// specifically, it's used in the transparent blits to determine if a pixel is
// transparent or not.
/*
int get_color(char *text, int bytes) {
  int c;
  switch(bytes) {
    case 1:
      __asm__ (
        "movb (%%eax), %%bl"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x000000FF;
    case 2:
      __asm__ (
        "movw (%%eax), %%bx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x0000ffff;
    case 3:
      __asm__ (
        "movl (%%eax), %%ebx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c & 0x00FFFFFF;
    case 4:
      __asm__ (
        "movl (%%eax), %%ebx"
        : "=b" (c)
        : "a" (text)
        : "memory"
      );
      return c;
  }
  return 0;
}
*/

void blit(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset;
  int pixel_height, pixel_width, width, i;
  int x1, y1, x2, y2, clipping;
  int tx1, tx2, t, c;     // used for transparent blits

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width of the
  // image goes off the screen.  Actually... it's easier,
  // and makes for less code, if we use the clipping code
  // if the image is transparent as well.
  clipping  = img->clipping | img->transparent | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // First we clip each image against themselves, using a standard
    // sutherland-hodgeman clipping algorithm
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Then calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // And a width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate the image and dislpay offsets (from the bottom left corner
    // because we'll work from bottom to top for speed)
    img_offset = (y1 +pixel_height) * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    // copy clipping image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      // decrement to next scanline
      img_offset -= img->bytes_per_line;
      dsp_offset -= petal.display.bytes_per_line;

      // And display it (using either the transparent or full method)
      if(img->transparent) {
        // these will hold non-transparent spans to copy...
        tx1 = tx2 = x1 * img->bytes_per_pixel;
        t = img->trans_col;         // the transparent colour which will not be copied
        while(tx1 < width) {
          // Try and find the first non-transparent pixel
          c = get_color(&img->address[img_offset + tx1], img->bytes_per_pixel);
          if(c != t) {
            tx2 = tx1;  // start at the beginning of the non-transparent span
            // and try and find the end of it...
            while(get_color(&img->address[img_offset + tx2], img->bytes_per_pixel) != t &&
                  tx2 < width)
              tx2 += img->bytes_per_pixel;  // we haven't found the end yet so try the next pixel
            // Now draw the span
            memcpy(&image->data[dsp_offset + tx1], &img->address[img_offset + tx1], tx2-tx1);
            tx1 = tx2;                     // search for the next span after this one, ofcourse
          }
          tx1 += img->bytes_per_pixel;    // goto the next pixel (haven't found start of span yet)
        }
      }
      else  // Not transparent, still clipped...
        memcpy(&image->data[dsp_offset], &img->address[img_offset], width);
    }
    // Now display the actual copied image
    XShmPutImage(display, window, gc, image, x, y, x, y, pixel_width, pixel_height, False);
  }

  else {       // No clipping needed
    // Copy the whole image in a couple steps when possible (no clipping needed)
    memcpy(image->data, img->address, img->size);
    XShmPutImage(display, window, gc, image, 0, 0, 0, 0, img->width, img->height, False);
  }

  // Now sync the screen (ie: force X11 to draw it)
  XSync(display, False);
  return;
}

void get_image(Image *img, signed int x, signed int y) {
  int dsp_offset, img_offset;
  int pixel_height, pixel_width, width, i;
  int x1, y1, x2, y2, clipping;

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width of the
  // image goes off the screen.
  clipping  = img->clipping | petal.display.clipping;
  clipping |= !(img->size == petal.display.size);
  clipping |= (img->width  + x + petal.display.x_off) > petal.display.width;
  clipping |= (img->height + y + petal.display.y_off) > petal.display.height;

  if(clipping) {
    // First we clip each image against themselves, using a standard
    // sutherland-hodgeman clipping algorithm
    if(!clip(img, &petal.display, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Then calculate image width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // and a width in bytes
    width = pixel_width * img->bytes_per_pixel;

    // Calculate the image and dislpay offsets (from the bottom left corner
    // because we'll work from bottom to top for speed)
    img_offset = (y1 +pixel_height) * img->bytes_per_line + (x1 * img->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * petal.display.bytes_per_line + (x * petal.display.bytes_per_pixel);

    // copy clipping image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      img_offset -= img->bytes_per_line;
      dsp_offset -= petal.display.bytes_per_line;
      memcpy(&img->address[img_offset], &image->data[dsp_offset], width);
    }
  }
  else {
    // Copy the whole image in a couple steps when possible (no clipping needed)
    memcpy(img->address, image->data, img->size);
  }

  // Now sync the screen (ie: force X11 to draw it)
  XSync(display, False);
  return;
}

char get_pointer(int *x, int *y) {
  while( XCheckTypedEvent(display, MotionNotify, &event) ) {
    px = event.xmotion.x;
    py = event.xmotion.y;
  }

  while(	XCheckTypedEvent(display, ButtonPress, &event) )
    button |= event.xbutton.button;

  while(	XCheckTypedEvent(display, ButtonRelease, &event) )
    button &= !event.xbutton.button;

  // These should never happen, but I'm checkin' for them anyway...
  if(px < 0) px = 0;
  if(py < 0) py = 0;
  if(px > petal.display.width) px = petal.display.width;
  if(py > petal.display.height) py = petal.display.height;

  *x = px; *y = py;
  return button;
}

char get_key(void) {
  char key = 0;

  while( XCheckTypedEvent(display, KeyPress, &event) )
    key = XKeycodeToKeysym(display, event.xkey.keycode, 0);

  return key;
}

void set_pal(rgb *pal) {
  int i;

  for(i = 0; i < 256; i++) {
    temp[i].pixel = i;
    temp[i].flags = DoRed | DoGreen | DoBlue;
    temp[i].red   = (int)pal[i].r * 256;
    temp[i].blue  = (int)pal[i].b * 256;
    temp[i].green = (int)pal[i].g * 256;
  }
  if(petal.pal != NULL) XFreeColormap(display, cmap);
  cmap=XCreateColormap(display, RootWindowOfScreen(screen),
                       DefaultVisualOfScreen(screen), AllocAll);
  XStoreColors(display, cmap, temp, 256);
  XSetWindowColormap(display,window,cmap);
}
