/* filled.cc
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * Contains filled versions of all functions and the screen clear function.
 * Each function in here now supports clipping and offseting.
 */

#include <iostream.h>
#include <stdlib.h>
#include "petal.h"
#include "filled.h"

void clear8(int c, Image *img = petal.default_image) {
  int colour = 0;
  int clipping;

  /* We'll use filled_rect under any of these circumstances.  The reason
   * for the width check is because we may be dealing with a sub image.
   */
  clipping  = img->x_off | img->y_off;
  clipping |= img->clipping;
  clipping |= (img->width != img->bytes_per_line);

  if(clipping) {
    filled_rect(0,0,img->width-1, img->height-1, c, img);
    return;
  }

  colour = (c << 24) |
           (c << 16) |
           (c << 8)  |
            c;

  /* This code is very similar to my blitting routine except that it
   * doesn't need to load eax with a colour... it's specified and always
   * the same
   */
  __asm__ ("
    0:
      movl %%eax, (%%edi,%%ebx,8);
      movl %%eax, 4(%%edi,%%ebx,8);
      inc  %%ebx;
      dec  %%ecx;
      jnz  0b;
    "
    :
    : "D"(img->address),"c"((img->size >> 3)),"b"(0),"a"(colour)
    : "bx", "cx"
  );
}

void clear16(int c, Image *img = petal.default_image) {
  int colour = 0;
  int clipping;

  /* We'll use filled_rect under any of these circumstances.  The reason
   * for the width check is because we may be dealing with a sub image.
   */
  clipping  = img->x_off | img->y_off;
  clipping |= img->clipping;
  clipping |= ((img->width<<1) != img->bytes_per_line);

  if(clipping) {
    filled_rect(0,0,img->width-1, img->height-1, c, img);
    return;
  }

  colour = (c << 16) |
            c;

  /* This code is very similar to my blitting routine except that it
   * doesn't need to load eax with a colour... it's specified and always
   * the same
   */
  __asm__ ("
    0:
      movl %%eax, (%%edi,%%ebx,8);
      movl %%eax, 4(%%edi,%%ebx,8);
      inc  %%ebx;
      dec  %%ecx;
      jnz  0b;
    "
    :
    : "D"(img->address),"c"((img->size >> 3)),"b"(0),"a"(colour)
    : "bx", "cx"
  );
}

/* crappy... needs to be fixed */
void clear24(int c, Image *img = petal.default_image) {
  int i, clipping;

  /* We'll use filled_rect under any of these circumstances.  The reason
   * for the width check is because we may be dealing with a sub image.
   */
  clipping  = img->x_off | img->y_off;
  clipping |= img->clipping;
  clipping |= ((img->width*3) != img->bytes_per_line);

  if(clipping) {
    filled_rect(0,0,img->width-1, img->height-1, c, img);
    return;
  }

  for(i = 0; i < img->height; i++) hline(0,img->width,i,c,img);
}

void clear32(int c, Image *img = petal.default_image) {
  int clipping;

  /* We'll use filled_rect under any of these circumstances.  The reason
   * for the width check is because we may be dealing with a sub image.
   */
  clipping  = img->x_off | img->y_off;
  clipping |= img->clipping;
  clipping |= ((img->width<<2) != img->bytes_per_line);

  if(clipping) {
    filled_rect(0,0,img->width-1, img->height-1, c, img);
    return;
  }

  /* This code is very similar to my blitting routine except that it
   * doesn't need to load eax with a colour... it's specified and always
   * the same
   */
  __asm__ ("
    0:
      movl %%eax, (%%edi,%%ebx,8);
      movl %%eax, 4(%%edi,%%ebx,8);
      inc  %%ebx;
      dec  %%ecx;
      jnz  0b;
    "
    :
    : "D"(img->address),"c"((img->size >> 3)),"b"(0),"a"(c)
    : "bx", "cx"
  );
}

void filled_rect8(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  int colour = 0;
  char *address = img->address;

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  if(y1 > y2) {                        /* switch our y's if needed */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }
  if(x1 > x2) {                        /* switch our x's if needed */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x1) x1 = img->cx1;
    if(img->cy1 > y1) y1 = img->cy1;
    if(img->cx2 < x2) x2 = img->cx2;
    if(img->cy2 < y2) y2 = img->cy2;
    if(x1 > x2) return;
    if(y1 > y2) return;
  }

  /* If we just have a sliver, then use a more appropriate function */
  if(x1 == x2) { vline(y1,y2,x1,c,img); return; }
  if(y1 == y2) { hline(x1,x2,y1,c,img); return; }

  /* calculate the offset and add it on */
  address += y1 * img->bytes_per_line + x1;

  colour = (c << 24) |
           (c << 16) |
           (c << 8)  |
            c;

  /* loop through all the scan lines and draw the actual line */
  for(int y = y1; y <= y2; y++) {
    asm __volatile__("
      testl $1, %%edi;    # test if we're on an odd boundary
      jz 0f;              # if not, do nothing
      stosb;              # if so, output a byte
      decl %%ecx;         # and adjust our count accordinly
      jz 2f;              # quit if that was the last byte

    0:                    # check_word:
      testl $2, %%edi;    # check if we're on a word boundard
      jz 1f;              # if not, do nothing
      stosb;              # if so, it'd be great to output a word, however
      decl %%ecx;         # to be pixel perfect we'll output two bytes testing
      jz 2f;              # between each one to see if the count is zero
      stosb;
      decl %%ecx;
      jz 2f;

    1:                    # dwords:
      movl %%ecx, %%ebx;  # Copy the count into ebx and get the first
      andl $3, %%ebx;     # two bits.  These will be the remaining bytes to
                          # output after our dword outputs.

      shrl $2, %%ecx;     # divide ecx by 4, therefore getting number of dwords
      rep; stosl;         # now output all our dwords

      movl %%ebx, %%ecx;  # get the remaining bytes
      rep; stosb;         # and output them

    2:                    # done:
      "
      :
      : "D" (address), "a" (colour), "c" ((int)(x2-x1+1))
      : "ebx", "ecx", "edi", "memory"
    );
    address += img->bytes_per_line;     /* goto the next line */
  }
}


void filled_rect16(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  int colour = 0;
  char *address = img->address;

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  if(y1 > y2) {                        /* switch our y's if needed */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }
  if(x1 > x2) {                        /* switch our x's if needed */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  /* Clip if we're told to */
  if(img->clipping) {
    if(x1 < img->cx1) x1 = img->cx1;
    if(y1 < img->cy1) y1 = img->cy1;
    if(x2 > img->cx2) x2 = img->cx2;
    if(y2 > img->cy2) y2 = img->cy2;
    if(x1 > x2) return;
    if(y1 > y2) return;
  }

  /* If we just have a sliver, then use a more appropriate function */
  if(x1 == x2) { vline(y1,y2,x1,c,img); return; }
  if(y1 == y2) { hline(x1,x2,y1,c,img); return; }

  /* calculate the offset and add it on */
  address += y1 * img->bytes_per_line + (x1 << 1);

  colour = (c << 16) |
            c;

  /* loop through all the scan lines and draw the actual line */
  for(int y = y1; y <= y2; y++) {
    asm __volatile__ ("
    0:                    # check_word:
      testl $2, %%edi;    # check if we're on a word boundary
      jz 1f;              # if not, do nothing
      stosw;              # if so, output a word
      subl $2, %%ecx;     # adjust ecx accordingly
      jz 2f;              # and jump if we're done

      # you'll notice from the above there's no need to output individual bytes
      # anymore either.  A word will suffice because pixels are larger than a byte.

    1:                    # dwords:
      movl %%ecx, %%ebx;  # Copy the count into ebx and get the first
      andl $3, %%ebx;     # two bits.  ebx can now be either 0b or 10b
                          # On the later, we'll ouput a word before exiting

      shrl $2, %%ecx;     # divide ecx by 4, therefore getting number of dwords
      rep; stosl;         # now output all our dwords

      orl  %%ebx, %%ebx;  # check for any remaining words
      jz 2f;              # if none, exit
      stosw;              # else output the remaining word

    2:                    # done:
      "
      :
      : "D" (address), "a" (colour), "c" ((int)((x2-x1)<<1)+2)
      : "ebx", "ecx", "edi", "memory"
    );
    address += img->bytes_per_line;     /* goto the next line */
  }
}

void filled_rect24(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  int eax, ebx, ecx;
  char *address = img->address;

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  if(y1 > y2) {                        /* switch our y's if needed */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }
  if(x1 > x2) {                        /* switch our x's if needed */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x1) x1 = img->cx1;
    if(img->cy1 > y1) y1 = img->cy1;
    if(img->cx2 < x2) x2 = img->cx2;
    if(img->cy2 < y2) y2 = img->cy2;
    if(x1 > x2) return;
    if(y1 > y2) return;
  }

  /* If we just have a sliver, then use a more appropriate function */
  if(x1 == x2) { vline(y1,y2,x1,c,img); return; }
  if(y1 == y2) { hline(x1,x2,y1,c,img); return; }

  /* calculate the offset and add it on */
  address += y1 * img->bytes_per_line + (x1 * 3);

  asm __volatile__ ("
    # Make the first dword (eax):
    andl $0x00FFFFFF,%%eax;        # Take out the top byte
    movl %%eax,%%ebx;              # copy remaining to ebx
    shll $24,%%eax;                # get the LSB to the MSB
    orl  %%ebx,%%eax;              # OR eax and ebx to get first dword

    # make the second dword (ebx)
    shrl $8,%%ebx;                 # get rid of the LSB of ebx
    movl %%eax,%%ecx;              # copy eax into ebx
    shll $16,%%ecx;                # get the LSW to the MSW
    orl  %%ecx,%%ebx;              # OR ecx and ebx to get second dword

    # make the third dword (ebx)
    movl %%eax,%%ecx;              # copy eax into ecx
    shll $8,%%ecx;                 # shift it left one byte
    movb %%bh,%%cl;                # one copy in the right byte
    "
    : "=a" (eax), "=b" (ebx), "=c" (ecx)
    : "a" (c)
    : "eax", "ebx", "ecx"
  );

  /* loop through all the scan lines and draw the actual line */
  for(int y = y1; y <= y2; y++) {
    asm __volatile__ ("
      movl %%edx,%%esi;              # save our count for later purposes

      shrl $2,%%edx;                 # divide by 4, thereby getting # dwords
      jz 1f;                         # jump to byte outputs if not dwords

    0:
      movl %%eax,(%%edi);            # copy first dword
      movl %%ebx,4(%%edi);           # copy second dword
      movl %%ecx,8(%%edi);           # copy third dword
      addl $12,%%edi;                # goto the next 12 bytes of the address
      decl %%edx;                    # and decrease our count
      jnz 0b;                        # loop until no more to go

    1:
      andl $3,%%esi;                 # find out how many single pixels left
      jz 3f;                         # jump out if finished

      andl $0x00FFFFFF,%%eax;        # get rid of top byte of first dword
    2:
      andl $0xFF000000,(%%edi);      # mask of 24-bit pixel in memory
      orl %%eax,(%%edi);             # OR it into place
      addl $3,%%edi;                 # goto the next pixel
      decl %%esi;                    # decrease our count
      jnz 2b;                        # and continue until done
    3:
      "
      :
      : "D" (address), "a" (eax), "b" (ebx), "c" (ecx), "d" ((int)(x2-x1+1))
      : "eax", "ebx", "ecx", "edx", "edi", "esi", "memory"
    );
    address += img->bytes_per_line;     /* goto the next line */
  }
}

void filled_rect32(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  char *address = img->address;

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  if(y1 > y2) {                        /* switch our y's if needed */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }
  if(x1 > x2) {                        /* switch our x's if needed */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x1) x1 = img->cx1;
    if(img->cy1 > y1) y1 = img->cy1;
    if(img->cx2 < x2) x2 = img->cx2;
    if(img->cy2 < y2) y2 = img->cy2;
    if(x1 > x2) return;
    if(y1 > y2) return;
  }

  /* If we just have a sliver, then use a more appropriate function */
  if(x1 == x2) { vline(y1,y2,x1,c); return; }
  if(y1 == y2) { hline(x1,x2,y1,c); return; }

  /* calculate the offset and add it on */
  address += y1 * img->bytes_per_line + (x1 << 2);

  /* loop through all the scan lines and draw the actual line */
  for(int y = y1; y <= y2; y++) {
    asm __volatile__("
      # No need for the odd or even test, because in 32bpp everything is dword aligned
      # All we have to do is output them!  Simple as that!
      rep; stosl;         # now output all our dwords
      "
      :
      : "D" (address), "a" (c), "c" ((int)(x2-x1+1))
      : "ebx", "ecx", "edi", "memory"
    );
    address += img->bytes_per_line;     /* goto the next line */
  }
}
