/* font.cc
 * 1997/12/27
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * Contains function to support BGI scalable fonts.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>

#include "petal.h"
#include "font.h"

char font_magic[4] = { 'P', 'K', 8, 8 };
char  bgi_magic[4] = { 'B', 'G', 'I', ' ' };

Font::Font(void) {
  numer = denom = 1;
  col = 0;    // black
  isbold = 1;
}

void Font::load(char *fname) {
  int i, font_size;
  FILE *file = fopen(fname, "rb");
  
  // Check for a valid file
  if(file == NULL) {
    cout << "Cannot load font: " << fname << endl;
    exit(1);
  }
  
  // Read in the header and check if it's valid
  fread(&header, 0x90, 1, file);
  if(strncmp((char *)header,font_magic,4)) {
    cout << "Invalid font file\n";
    exit(1);
  }
  if(strncmp((char *)&header[4],bgi_magic,4)) {
    cout << "Invalid font file\n";
    exit(1);
  }

  i = 8;
  while(header[i] != 26) { /*cout << header[i];*/ i++; }

  // The next word in the font will be the header size, but I'll just
  // assume 0x80.  Every font I've run across has a header size of 0x80.

  // After that is a 4 character internat font name.  We, ofcourse, don't
  // care about this, so we'll skip over it.  After that is the size of the
  // font.  We'll need to save this for later use, adjusting it to remove the 16
  // byte "secondary header."
  i += 7;
  font_size = (unsigned int)header[i+1]; font_size <<= 8; font_size += (unsigned int)header[i];
  font_size -= 0x10;
  //printf("Font size: %d\n", font_size);

  //i += 2; printf("Font version: %x.%x\n", header[i], header[i+1]);

  // Now we're past the main header (the first 0x80 bytes) and in the secondary
  // header.  Everything here is stationary so we no longer need i to track location.
  // Again, we'll adjust our font size to remove the width, and offset tables of the font.
  if(header[0x80] != '+') { cout << "Font Not stroked\n"; exit(1); }
  chars = (unsigned int)header[0x82]; chars <<= 8; chars += (unsigned int)header[0x81];
  font_size -= chars*3;
  //printf("%d characters in font\n", chars);

  first_char = header[0x84];
  //printf("First character: %d\n", first_char);
  to_top = header[0x88];
  to_bottom = header[0x89];
  to_descender = header[0x8A];
  //printf("Font metrics: %d %d %d\n", to_top, to_bottom, to_descender);

  // Now we'll allocate room for the character definition offsets
  // and the character width tables.  Then simply read them in.
  font_offset = new short[chars];
  font_width = new char[chars];

  fread(font_offset, chars*2, 1, file);
  fread(font_width, chars, 1, file);

  // Now read in the rest of the file, which is the actual font description data
  font_data = new char[font_size];
  fread(font_data, font_size, 1, file);

  // And we're done
  fclose(file);
}

void Font::outtextxy(int tx, int ty, char *string, Image *img = petal.default_image, int len = 0) {
  unsigned int  i, byte;
  signed char xcode, ycode, opcode;
  signed int  x_loc, y_loc;

  // Loop for all characters in the string
  for(i = 0; i < strlen(string); i++) {
    // get the offset of the character definition
    byte = font_offset[string[i] - first_char];
    // And loop until we get the END opcode
    for( ; ; ) {
      // get our x and y codes and increment to the next code
      xcode = font_data[byte];
      ycode = font_data[byte+1];
      byte += 2;

      // Extract the opcode and x, y coordinates from the codes
      opcode = ((xcode & 128) >> 7) | ((ycode & 128) >> 6);
      xcode &= 127; ycode &= 127;
      // If xcode is negative we must infill the character to 8-bits from 7
      if(xcode & 64) xcode |= 128;
      if(ycode & 64) ycode |= 128;
      // Scale the font location coordinate
      x_loc = xcode * numer / denom;
      y_loc = ycode * numer / denom;

      // If we've gotten the END opcode, then exit the loop
      if(opcode == 0) break;

      // Else do what the opcode tells us to
      switch(opcode) {
        // Set current position opcode...
        case 1:
          x = x_loc;
          y = y_loc;
          break;
        // Scan opcode.  I have no idea what to do here!
        case 2:
          cout << "Scanning on character " << string[i] << endl;
          break;
        // Draw to position opcode...
        case 3:
          //line(tx + x, ty - y, tx + x_loc, ty - y_loc, col, img);
          thick_line(tx + x, ty - y, tx + x_loc, ty - y_loc, col, isbold * numer / denom, img);
          x = x_loc;
          y = y_loc;
          break;
      }
    }
    // Increase of x location to get to the next character's spot
    tx += (int)(font_width[string[i] - first_char] * numer / denom) + isbold;
    if(len != 0 && tx >= len) break;
  }
  return;
}

int Font::length(char *string) {
  unsigned int i, len = 0;

  for(i = 0; i < strlen(string); i++)
    len += (int)(font_width[string[i] - first_char] * numer / denom) + isbold;

  return len;
}

void Font::thick_line(int x1, int y1, int x2, int y2, int c, int thickness, Image *img = petal.default_image) {
  int half, i;

  if(thickness == 0) thickness = 1;
  half = -(thickness >> 1);

  // Make sure the thick line is centered...
  for(i = half; i < (half + thickness); i++) line(x1 + i, y1, x2 + i, y2, c, img);
}
