/* lines.cc
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * Contains different portable line drawing functions for PETAL.
 */

#include <iostream.h>
#include <stdlib.h>
#include "petal.h"
#include "lines.h"

/* some clipping variables and routines */
enum { left   = 8,
       right  = 4,
       below  = 2,
       above  = 1,
       inside = 0 };
typedef int outcode;
    
outcode getOut(int x, int y, Image *img) {
  outcode o = inside;

  if ( x < img->cx1 )      o |= left;
  else if ( x > img->cx2 ) o |= right;
  if ( y < img->cy1 )      o |= above;
  else if ( y > img->cy2 ) o |= below;
   
  return o;
}

int clipLine(int *x1, int *y1, int *x2, int *y2, Image *img) {
  outcode A, B, P;                                  /* P is used to hold either A or B */
  int x = 0, y = 0, Point_is_A;                     /* x and y hold clipping values, and Point_is_A tells which point P holds */

  A = getOut(*x1,*y1,img); B = getOut(*x2,*y2,img); /* get the outcodes for point A and B of the line */
  if( (A + B) == 0) return 0;                       /* no clipping necessary */
  if( (A & B) != 0) return 1;                       /* invisible */

  while(A || B)  {                                  /* loop while there's still clipping to be done */
    if( (A & B) != 0) return 1;                     /* but exit if it's realized that the line is really invisible */

    if ( A )  { P = A; Point_is_A = 1; }            /* decide which point to clip this time around */
    else      { P = B; Point_is_A = 0; }

    /* Clip to each clipping plane */
    if ( P & left )  {
      y = (img->cx1 - *x1) * (*y2 - *y1)/(*x2 - *x1) + *y1;
      x = img->cx1;
    }
    else if ( P & right )  {
      y = (img->cx2 - *x1) * (*y2 - *y1)/(*x2 - *x1) + *y1;
      x = img->cx2;
    }
    if ( P & above )  {
      x = (img->cy1 - *y1) * (*x2 - *x1)/(*y2 - *y1) + *x1;
      y = img->cy1;
    }
    else if ( P & below )  {
      x = (img->cy2 - *y1) * (*x2 - *x1)/(*y2 - *y1) + *x1;
      y = img->cy2;
    }

    /* check to see what the new outcodes of P are */
    P = getOut(x,y,img);
    /* and assign x, y and the new outcodes back to the right point */
    if ( Point_is_A )  { *x1 = x; *y1 = y; A = P; }
    else  { *x2 = x; *y2 = y; B = P; }

    /* It's possible that a horizontal line will be created by clipping a very
     * flat line.  This line cannot be clipped using the above clipping
     * technique (divide by zero in the slope calculation).  A vertical line
     * probably wont work either.  In each of these situations the clipping
     * function will exit.  It's the job of the calling function to call the
     * hline or vline function if then needed.
     */
    if(*x1 == *x2) { return 0; }
    if(*y1 == *y2) { return 0; }
  }
  return 0;
}


/* -- REGULAR LINES ---- */
void line8(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  int dx, dy, xi, yi, error = 0, lcv, offset;

  /* Plot a horizontal or vertical line when possible */
  if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
  if(y1 == y2) { hline(x1, x2, y1, c, img); return; }

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  /* do some clipping */
  if(img->clipping) {
    if(clipLine(&x1, &y1, &x2, &y2, img) == 1) return;

    /* read the comment for the clipLine function to see why this is necessary */
    if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
    if(y1 == y2) { hline(x1, x2, y1, c, img); return; }
  }

  offset = y1 * img->bytes_per_line + x1;

  if(x2 >= x1) xi =  1;
  else         xi = -1;
  if(y2 >= y1) yi =  img->bytes_per_line;
  else         yi = -img->bytes_per_line;

  dx = abs(x2-x1) + 1;
  dy = abs(y2-y1) + 1;

  if(dx > dy) {
    for(lcv = 0; lcv < dx; lcv++) {
      __asm__ (
        "movb %%bl, (%%eax)"
        :
        : "a" (img->address + offset), "b" ((char)c)
        : "memory"
      );
      error += dy;
      if(error > dx) {
        error -= dx;
        offset += yi;
      }
      offset += xi;
    }
  }
  else {
    for(lcv = 0; lcv < dy; lcv++) {
      __asm__ (
        "movb %%bl, (%%eax)"
        :
        : "a" (img->address + offset), "b" (c)
        : "memory"
      );
      error += dx;
      if(error > 0) {
        error -= dy;
        offset += xi;
      }
      offset += yi;
    }
  }
  return;
}


void line16(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  int dx, dy, xi, yi, error = 0, lcv, offset;

  /* Plot a horizontal or vertical line when possible */
  if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
  if(y1 == y2) { hline(x1, x2, y1, c, img); return; }

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  /* do some clipping */
  if(img->clipping) {
    if(clipLine(&x1, &y1, &x2, &y2, img) == 1) return;

    /* read the comment for the clipLine function to see why this is necessary */
    if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
    if(y1 == y2) { hline(x1, x2, y1, c, img); return; }
  }

  offset = y1 * img->bytes_per_line + (x1 << 1);

  if(x2 >= x1) xi =  2;
  else         xi = -2;
  if(y2 >= y1) yi =  img->bytes_per_line;
  else         yi = -img->bytes_per_line;

  dx = abs(x2-x1) + 1;
  dy = abs(y2-y1) + 1;

  if(dx > dy) {
    for(lcv = 0; lcv < dx; lcv++) {
      __asm__ (
        "movw %%bx, (%%eax)"
        :
        : "a" (img->address + offset), "b" (c)
        : "memory"

      );
      error += dy;
      if(error > dx) {
        error -= dx;
        offset += yi;
      }
      offset += xi;
    }
  }
  else {
    for(lcv = 0; lcv < dy; lcv++) {
      __asm__ (
        "movw %%bx, (%%eax)"
        :
        : "a" (img->address + offset), "b" (c)
        : "memory"
      );
      error += dx;
      if(error > 0) {
        error -= dy;
        offset += xi;
      }
      offset += yi;
    }
  }
  return;
}


void line24(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  int dx, dy, xi, yi, error = 0, lcv, offset;

  /* Plot a horizontal or vertical line when possible */
  if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
  if(y1 == y2) { hline(x1, x2, y1, c, img); return; }

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  /* do some clipping */
  if(img->clipping) {
    if(clipLine(&x1, &y1, &x2, &y2, img) == 1) return;

    /* read the comment for the clipLine function to see why this is necessary */
    if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
    if(y1 == y2) { hline(x1, x2, y1, c, img); return; }
  }

  offset = y1 * img->bytes_per_line + (x1 * 3);

  if(x2 >= x1) xi =  3;
  else         xi = -3;
  if(y2 >= y1) yi =  img->bytes_per_line;
  else         yi = -img->bytes_per_line;

  dx = abs(x2-x1) + 1;
  dy = abs(y2-y1) + 1;

  if(dx > dy) {
    for(lcv = 0; lcv < dx; lcv++) {
      __asm__ (
       "andl $0x00FFFFFF,%%ebx;     # mask off the top byte of the colour
        andl $0xFF000000,(%%eax);   # mask off the actual 24-bit value
        orl  %%ebx,(%%eax);         # and or it into place
       "
        :
        : "a" (img->address + offset), "b" (c)
        : "memory"
      );
      error += dy;
      if(error > dx) {
        error -= dx;
        offset += yi;
      }
      offset += xi;
    }
  }
  else {
    for(lcv = 0; lcv < dy; lcv++) {
      __asm__ (
       "andl $0x00FFFFFF,%%ebx;     # mask off the top byte of the colour
        andl $0xFF000000,(%%eax);   # mask off the actual 24-bit value
        orl  %%ebx,(%%eax);         # and or it into place
       "
        :
        : "a" (img->address + offset), "b" (c)
        : "memory"
      );
      error += dx;
      if(error > 0) {
        error -= dy;
        offset += xi;
      }
      offset += yi;
    }
  }
  return;
}


void line32(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  int dx, dy, xi, yi, error = 0, lcv, offset;

  /* Plot a horizontal or vertical line when possible */
  if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
  if(y1 == y2) { hline(x1, x2, y1, c, img); return; }

  OffsetAdjust(img, x1, y1);
  OffsetAdjust(img, x2, y2);

  /* do some clipping */
  if(img->clipping) {
    if(clipLine(&x1, &y1, &x2, &y2, img) == 1) return;

    /* read the comment for the clipLine function to see why this is necessary */
    if(x1 == x2) { vline(y1, y2, x1, c, img); return; }
    if(y1 == y2) { hline(x1, x2, y1, c, img); return; }
  }

  offset = y1 * img->bytes_per_line + (x1 << 2);

  if(x2 >= x1) xi =  4;
  else         xi = -4;
  if(y2 >= y1) yi =  img->bytes_per_line;
  else         yi = -img->bytes_per_line;

  dx = abs(x2-x1) + 1;
  dy = abs(y2-y1) + 1;

  if(dx > dy) {
    for(lcv = 0; lcv < dx; lcv++) {
      __asm__ (
        "movl %%ebx, (%%eax)"
        :
        : "a" (img->address + offset), "b" (c)
        : "memory"
      );
      error += dy;
      if(error > dx) {
        error -= dx;
        offset += yi;
      }
      offset += xi;
    }
  }
  else {
    for(lcv = 0; lcv < dy; lcv++) {
      __asm__ (
        "movl %%ebx, (%%eax)"
        :
        : "a" (img->address + offset), "b" (c)
        : "memory"
      );
      error += dx;
      if(error > 0) {
        error -= dy;
        offset += xi;
      }
      offset += yi;
    }
  }

  return;
}


/* -- HORIZONTAL LINES ---- */
void hline8(int x1, int x2, int y, int c, Image *img = petal.default_image) {
  int colour = 0;
  char *address = img->address;

  /* plot a single pixel when possible */
  //if(x1 == x2) putpixel(x1, y, c, img);

  OffsetAdjust(img, x1, y);
  x2 += img->x_off;

  if(x1 > x2) {                        /* swap if necessary */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  if(img->clipping) {
    if(y > img->cy2) return;
    if(y < img->cy1) return;
    if(x1 < img->cx1) x1 = img->cx1;
    if(x2 > img->cx2) x2 = img->cx2;
    if(x1 > x2) return;                 /* Nothing to draw... */
  }

  /* we have to do this here, after we swap the x's, and perform clipping. */
  address += y * img->bytes_per_line + x1;

  /* Make a dword of our 8 bit colour */
  colour = (c << 24) |
           (c << 16) |
           (c << 8)  |
            c;

  /* plot a single pixel if we must */
  if(x1 == x2) { fputpixel(address, c); return; }

  __asm__ __volatile__("
    testl $1, %%edi;    # test if we're on an odd boundary
    jz 0f;              # if not, do nothing
    stosb;              # if so, output a byte
    decl %%ecx;         # and adjust our count accordinly
    jz 2f;              # quit if that was the last byte

  0:                    # check_word:
    testl $2, %%edi;    # check if we're on a word boundard
    jz 1f;              # if not, do nothing
    stosb;              # if so, it'd be great to output a word, however
    decl %%ecx;         # to be pixel perfect we'll output two bytes testing
    jz 2f;              # between each one to see if the count is zero
    stosb;
    decl %%ecx;
    jz 2f;

  1:                    # dwords:
    movl %%ecx, %%ebx;  # Copy the count into ebx and get the first
    andl $3, %%ebx;     # two bits.  These will be the remaining bytes to
                        # output after our dword outputs.

    shrl $2, %%ecx;     # divide ecx by 4, therefore getting number of dwords
    rep; stosl;         # now output all our dwords

    movl %%ebx, %%ecx;  # get the remaining bytes
    rep; stosb;         # and output them

  2:                    # done:
    "
    :
    : "D" (address), "a" (colour), "c" ((int)(x2-x1))
    : "ebx", "ecx", "edi", "memory"
  );
}

void hline16(int x1, int x2, int y, int c, Image *img = petal.default_image) {
  int colour = 0;
  char *address = img->address;

  /* plot a single pixel when possible */
  //if(x1 == x2) putpixel(x1, y, c, img);

  OffsetAdjust(img, x1, y);
  x2 += img->x_off;

  if(x1 > x2) {                        /* swap if necessary */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  if(img->clipping) {
    if(y > img->cy2) return;
    if(y < img->cy1) return;
    if(x1 < img->cx1) x1 = img->cx1;
    if(x2 > img->cx2) x2 = img->cx2;
    if(x1 > x2) return;                 /* Nothing to draw... */
  }

  /* we have to do this here, after we swap the x's, and perform clipping. */
  address += y * img->bytes_per_line + (x1 << 1);

  /* Make a dword of our 16 bit colour */
  colour = (c << 16) | c;

  /* plot a single pixel if we must */
  if(x1 == x2) { fputpixel(address, c); return; }

  __asm__ __volatile__("
    # No need for the odd test, because in 16bpp everything is word aligned
  0:                    # check_word:
    testl $2, %%edi;    # check if we're on a word boundary
    jz 1f;              # if not, do nothing
    stosw;              # if so, output a word
    subl $2, %%ecx;     # adjust ecx accordingly
    jz 2f;              # and jump if we're done

    # you'll notice from the above there's no need to output individual bytes
    # anymore either.  A word will suffice because pixels are larger than a byte.

  1:                    # dwords:
    movl %%ecx, %%ebx;  # Copy the count into ebx and get the first
    andl $3, %%ebx;     # two bits.  ebx can now be either 0b or 10b
                        # On the later, we'll ouput a word before exiting

    shrl $2, %%ecx;     # divide ecx by 4, therefore getting number of dwords
    rep; stosl;         # now output all our dwords

    orl  %%ebx, %%ebx;  # check for any remaining words
    jz 2f;              # if none, exit
    stosw;              # else output the remaining word

  2:                    # done:
    "
    :
    : "D" (address), "a" (colour), "c" ((int)((x2-x1)<<1))
    : "ebx", "ecx", "edi", "memory"
  );
}

/* The 24bpp hline function.  This one's gotta change soon because it's not dword aligned
 * It's possible that this function isn't even needed if 24bpp modes are always 4 bytes,
 * in which case my 32bpp hline will work fine.
 */
void hline24(int x1, int x2, int y, int c, Image *img = petal.default_image) {
  char *address = img->address;

  /* plot a single pixel when possible */
  //if(x1 == x2) putpixel(x1, y, c, img);

  OffsetAdjust(img, x1, y);
  x2 += img->x_off;

  if(x1 > x2) {                        /* swap if necessary */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  if(img->clipping) {
    if(y > img->cy2) return;
    if(y < img->cy1) return;
    if(x1 < img->cx1) x1 = img->cx1;
    if(x2 > img->cx2) x2 = img->cx2;
    if(x1 > x2) return;                 /* Nothing to draw... */
  }

  /* we have to do this here, after we swap the x's, and perform clipping. */
  address += y * img->bytes_per_line + (x1 * 3);

  /* plot a single pixel if we must */
  if(x1 == x2) { fputpixel(address, c); return; }

  asm __volatile__ ("
    # Make the first dword (eax):
    andl $0x00FFFFFF,%%eax;        # Take out the top byte
    movl %%eax,%%ebx;              # copy remaining to ebx
    shll $24,%%eax;                # get the LSB to the MSB
    orl  %%ebx,%%eax;              # OR eax and ebx to get first dword

    # make the second dword (ebx)
    shrl $8,%%ebx;                 # get rid of the LSB of ebx
    movl %%eax,%%ecx;              # copy eax into ebx
    shll $16,%%ecx;                # get the LSW to the MSW
    orl  %%ecx,%%ebx;              # OR ecx and ebx to get second dword

    # make the third dword (ebx)
    movl %%eax,%%ecx;              # copy eax into ecx
    shll $8,%%ecx;                 # shift it left one byte
    movb %%bh,%%cl;                # one copy in the right byte

    movl %%edx,%%esi;              # save our count for later purposes

    shrl $2,%%edx;                 # divide by 4, thereby getting # dwords
    jz 1f;                         # jump to byte outputs if not dwords

  0:
    movl %%eax,(%%edi);            # copy first dword
    movl %%ebx,4(%%edi);           # copy second dword
    movl %%ecx,8(%%edi);           # copy third dword
    addl $12,%%edi;                # goto the next 12 bytes of the address
    decl %%edx;                    # and decrease our count
    jnz 0b;                        # loop until no more to go

  1:
    andl $3,%%esi;                 # find out how many single pixels left
    jz 3f;                         # jump out if finished

    andl $0x00FFFFFF,%%eax;        # get rid of top byte of first dword
  2:
    andl $0xFF000000,(%%edi);      # mask of 24-bit pixel in memory
    orl %%eax,(%%edi);             # OR it into place
    addl $3,%%edi;                 # goto the next pixel
    decl %%esi;                    # decrease our count
    jnz 2b;                        # and continue until done
  3:
    "
    :
    : "D" (address), "a" (c), "d" ((int)(x2-x1))
    : "eax", "ebx", "ecx", "edx", "edi", "esi", "memory"
  );
}

void hline32(int x1, int x2, int y, int c, Image *img = petal.default_image) {
  char *address = img->address;

  /* plot a single pixel when possible */
  //if(x1 == x2) putpixel(x1, y, c, img);

  OffsetAdjust(img, x1, y);
  x2 += img->x_off;

  if(x1 > x2) {                        /* swap if necessary */
    x1 ^= x2;
    x2 ^= x1;
    x1 ^= x2;
  }

  if(img->clipping) {
    if(y > img->cy2) return;
    if(y < img->cy1) return;
    if(x1 < img->cx1) x1 = img->cx1;
    if(x2 > img->cx2) x2 = img->cx2;
    if(x1 > x2) return;                 /* Nothing to draw... */
  }

  /* we have to do this here, after we swap the x's, and perform clipping. */
  address += y * img->bytes_per_line + (x1 << 2);

  /* plot a single pixel if we must */
  if(x1 == x2) { fputpixel(address, c); return; }

  __asm__ __volatile__("
    # No need for the odd or even test, because in 32bpp everything is dword aligned
    # All we have to do is output them!  Simple as that!
    rep; stosl;         # now output all our dwords
    "
    :
    : "D" (address), "a" (c), "c" ((int)(x2-x1))
    : "ebx", "ecx", "edi", "memory"
  );
}

/* -- VERTICAL LINES ---- */
void vline8(int y1, int y2, int x, int c, Image *img = petal.default_image) {
  int offset, height, i;
  char *address = img->address;

  /* plot a single pixel when possible */
  if(y1 == y2) putpixel(x, y1, c, img);

  OffsetAdjust(img, x, y1);
  y2 += img->y_off;

  if(y1 > y2) {      /* swap if necessary */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }

  if(img->clipping) {
    if(x > img->cx2) return;
    if(x < img->cx1) return;
    if(y1 < img->cy1) y1 = img->cy1;
    if(y2 > img->cy2) y2 = img->cy2;
  }

  /* calculate the first address and the height of the line */
  offset = y1 * img->bytes_per_line + x;
  height = (y2 - y1);

  /* loop for the height of the line and plot pixels along the way */
  for(i = height; i >= 0; i--) {
    __asm__ ("movb %%bl, (%%eax)"
      :
      : "a" (address + offset), "b" (c)
      : "memory"
    );
    address += img->bytes_per_line;          /* goto the next line */
  }
}

void vline16(int y1, int y2, int x, int c, Image *img = petal.default_image) {
  int offset, height, i;
  char *address = img->address;

  /* plot a single pixel when possible */
  if(y1 == y2) putpixel(x, y1, c, img);

  OffsetAdjust(img, x, y1);
  y2 += img->y_off;

  if(y1 > y2) {      /* swap if necessary */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }

  if(img->clipping) {
    if(x > img->cx2) return;
    if(x < img->cx1) return;
    if(y1 < img->cy1) y1 = img->cy1;
    if(y2 > img->cy2) y2 = img->cy2;
  }

  /* calculate the first address and the height of the line */
  offset = (y1 * img->bytes_per_line)+(x << 1);
  height = (y2 - y1);

  /* loop for the height of the line and plot pixels along the way */
  for(i = height; i >= 0; i--) {
    __asm__ ("movw %%bx, (%%eax)"
      :
      : "a" (address + offset), "b" (c)
      : "memory"
    );
    address += img->bytes_per_line;          /* goto the next line */
  }
}

void vline24(int y1, int y2, int x, int c, Image *img = petal.default_image) {
  int offset, height, i;
  char *address = img->address;

  /* plot a single pixel when possible */
  if(y1 == y2) putpixel(x, y1, c, img);

  OffsetAdjust(img, x, y1);
  y2 += img->y_off;

  if(y1 > y2) {      /* swap if necessary */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }

  if(img->clipping) {
    if(x > img->cx2) return;
    if(x < img->cx1) return;
    if(y1 < img->cy1) y1 = img->cy1;
    if(y2 > img->cy2) y2 = img->cy2;
  }

  /* calculate the first address and the height of the line */
  offset = (y1 * img->bytes_per_line)+(x * 3);
  height = (y2 - y1);

  /* loop for the height of the line and plot pixels along the way */
  for(i = height; i >= 0; i--) {
    asm __volatile__ (
     "andl $0x00FFFFFF,%%ebx;     # mask off the top byte of the colour
      andl $0xFF000000,(%%eax);   # mask off the actual 24-bit value
      orl  %%ebx,(%%eax);         # and or it into place
     "
      :
      : "a" (address + offset), "b" (c)
      : "memory"
    );
    address += img->bytes_per_line;          /* goto the next line */
  }
}

void vline32(int y1, int y2, int x, int c, Image *img = petal.default_image) {
  int offset, height, i;
  char *address = img->address;

  /* plot a single pixel when possible */
  if(y1 == y2) putpixel(x, y1, c, img);

  OffsetAdjust(img, x, y1);
  y2 += img->y_off;

  if(y1 > y2) {      /* swap if necessary */
    y1 ^= y2;
    y2 ^= y1;
    y1 ^= y2;
  }

  if(img->clipping) {
    if(x > img->cx2) return;
    if(x < img->cx1) return;
    if(y1 < img->cy1) y1 = img->cy1;
    if(y2 > img->cy2) y2 = img->cy2;
  }

  /* calculate the first address and the height of the line */
  offset = (y1 * img->bytes_per_line)+(x << 2);
  height = (y2 - y1);

  /* loop for the height of the line and plot pixels along the way */
  for(i = height; i >= 0; i--) {
    __asm__ ("movl %%ebx, (%%eax)"
      :
      : "a" (address + offset), "b" (c)
      : "memory"
    );
    address += img->bytes_per_line;          /* goto the next line */
  }
}

void rect(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image) {
  hline(x1, x2, y1, c, img);
  hline(x1, x2, y2, c, img);
  vline(y1, y2, x1, c, img);
  vline(y1, y2, x2, c, img);
  return;
}
