/* pcx.cc
 * 1997/12/27
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * Contains functions to import PCX files into Images.  PETAL is able to load
 * 256 colour PCXs and convert them to any high colour depth if need be.  It
 * currently _will not_ load in high colour depth PCXs.  I doubt this will ever
 * be much of a concern, especially for game programers.  256 colours per texture
 * is enough, especially sence each texture could have different 256 colours.
 */

#include <iostream.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "petal.h"
#include "pcx.h"

Pcx::Pcx(void) {
  // Does nothing
}

Pcx::Pcx(char *filename) {
  load(filename);
}

Pcx::~Pcx(void) {
  if(img) delete_image(img);
}

// Load in a (256 colour) PCX
int Pcx::load(char *fname) {
  // Open the PCX file (return an error if the file doesn't exist)
  if ((file=fopen(fname, "rb"))==NULL) return 0;

  // store the filename (why, I don't know, but it's there.  I think my original
  // idea was so that I didn't load the same texture twice.  I implement such a
  // system in my 3D library; Benoit).
  strcpy((char *)filename, fname);

  // Read the file header and return an error if the colour depth isn't correct
  if(!loadheader()) return 0;

  // setup some quick reference variables (height and width)
  //width  = info->header.xMax - info->header.xMin + 1;
  //height = info->header.yMax - info->header.yMin + 1;
  //width = header.xMax+2;
  //height = header.yMax;

  // read the palette data.  We _must_ load the palette before the image
  // because I need the palette data to convert 256 colour images to
  // high colour
  loadpalette();

  // Read the image data, and convert where necessary
  loadimage();

  // finally, close PCX file
  fclose(file);
  return 1;
}

// Load the PCX into an Image
void Pcx::loadimage(void) {
  char *line, *buffer;
  int row, byte_width;
  int bytes;

  width  = header.xMax - header.xMin + 1;
  height = header.yMax - header.yMin + 1;
  bytes  = header.bytesPerLine * petal.display.bytes_per_pixel;
  line   = (char *)malloc(bytes);
  img    = new_image(width, height);

  // skip past header
  fseek(file, 128L, SEEK_SET);

  buffer = img->address;
  byte_width = width * img->bytes_per_pixel;

  for(row = 0; row < height; buffer += byte_width, ++row) {
    readline(line, header.bytesPerLine);
    memcpy(buffer, line, byte_width);
  }
}

void Pcx::readline(char *buffer, int bytes) {
  static unsigned char count = 0, value = 0;

  //if(header.encoding) {
    while(bytes--) {
      if(count == 0) {
        value = fgetc(file);
        if(value < 0xc0) {
          count= 1;
        } else {
          count = value - 0xc0;
          value = fgetc(file);
        }
      }
      count--;
      // store the byte in our image.  If need be I convert it to the proper colour
      // depth by knowing the maximum colour value in a 256 colour PCX is 256.
      if(header.bitsPerPixel == 8 && petal.display.bpp != 8)
         fputpixel(buffer, colour( palette[value].r / 255.0,
                                   palette[value].g / 255.0,
                                   palette[value].b / 255.0));
      else fputpixel(buffer, value);
      buffer += petal.display.bytes_per_pixel;
      //*(buffer++)= value;
    }
  //} else {
  //  fread(buffer, bytes, 1, file);
  //}
}


// Load in the pcx header
int Pcx::loadheader(void) {
  fseek(file, 0L, SEEK_SET);
  fread((char *)&header, sizeof(PcxHeader), 1, file);
  // Check that we've got the right colour depth or atleast one I can easily convert
  if(header.bitsPerPixel != 8 && header.bitsPerPixel != petal.display.bpp) {
    cout << "Wrong colour depth! " << header.bitsPerPixel << endl;
    return 0;
  }
  return 1;
}

// Load the PCX palette
void Pcx::loadpalette(void) {
  //unsigned int i;

  fseek(file, -768L, SEEK_END);
  fread(palette, 768, 1, file);

  /*
  for (i=0; i<256; i++) {
    // We must shift to the right by 2 to get the colour value (0 to 63) in the
    // bottom part of the byte.  Annoyingly, it's always in the top part of the
    // byte.  Why?  Who the hell knows?  It seems rather odd to me, anyway.

    // Actually, now that the palette code expects a number between 0 and 255
    // the shift is no longer needed.
    palette[i].r = fgetc(file);
    palette[i].g = fgetc(file);
    palette[i].b = fgetc(file);
  }
  */
}
