/* pixels.cc
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * Contains different portable pixel drawing functions for PETAL.  Each
 * of these functions now support clipping and offsetting.
 */

#include <iostream.h>
#include <stdlib.h>
#include "petal.h"
#include "pixels.h"

void putpixel8(int x, int y, int c, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return;
    if(img->cx2 < x || img->cy2 < y) return;
  }

  int offset = y * img->bytes_per_line + x;

  __asm__ (
    "movb %%bl, (%%eax)"
    :
    : "a" (img->address + offset), "b" (c)
    : "memory"
  );
}

void putpixel16(int x, int y, int c, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return;
    if(img->cx2 < x || img->cy2 < y) return;
  }

  int offset = y * img->bytes_per_line + (x << 1);

  __asm__ (
    "movw %%bx, (%%eax)"
    :
    : "a" (img->address + offset), "b" (c)
    : "memory"
  );
}

void putpixel24(int x, int y, int c, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return;
    if(img->cx2 < x || img->cy2 < y) return;
  }

  int offset = y * img->bytes_per_line + (x * 3);

  __asm__ (
   "andl $0x00FFFFFF,%%ebx;     # mask off the top byte of the colour
    andl $0xFF000000,(%%eax);   # mask off the actual 24-bit value
    orl  %%ebx,(%%eax);         # and or it into place
   "
    :
    : "a" (img->address + offset), "b" (c)
    : "memory"
  );
}

void putpixel32(int x, int y, int c, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return;
    if(img->cx2 < x || img->cy2 < y) return;
  }

  int offset = y * img->bytes_per_line + (x << 2);

  __asm__ (
    "movl %%ebx, (%%eax)"
    :
    : "a" (img->address + offset), "b" (c)
    : "memory"
  );
}

int getpixel8(int x, int y, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return 0;
    if(img->cx2 < x || img->cy2 < y) return 0;
  }

  int offset = y * img->bytes_per_line + x;
  int c;

  __asm__ (
    "movb (%%eax), %%bl"
    : "=b" (c)
    : "a" (img->address + offset)
    : "memory"
  );

  return c & 0x000000FF;
}

int getpixel16(int x, int y, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return 0;
    if(img->cx2 < x || img->cy2 < y) return 0;
  }

  int offset = y * img->bytes_per_line + (x << 1);
  int c;

  __asm__ (
    "movw (%%eax), %%bx"
    : "=b" (c)
    : "a" (img->address + offset)
    : "memory"
  );

  return c & 0x0000ffff;
}

int getpixel24(int x, int y, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return 0;
    if(img->cx2 < x || img->cy2 < y) return 0;
  }

  int offset = y * img->bytes_per_line + (x * 3);
  int c;

  __asm__ (
    "movl (%%eax), %%ebx"
    : "=b" (c)
    : "a" (img->address + offset)
    : "memory"
  );

  return c & 0x00FFFFFF;
}

int getpixel32(int x, int y, Image *img = petal.default_image) {
  OffsetAdjust(img, x, y);

  /* Clip if we're told to */
  if(img->clipping) {
    if(img->cx1 > x || img->cy1 > y) return 0;
    if(img->cx2 < x || img->cy2 < y) return 0;
  }

  int offset = y * img->bytes_per_line + (x << 2);
  int c;

  __asm__ (
    "movl (%%eax), %%ebx"
    : "=b" (c)
    : "a" (img->address + offset)
    : "memory"
  );

  return c;
}

void fputpixel8(char *offset, int c) {
  __asm__ (
    "movb %%bl, (%%eax)"
    :
    : "a" (offset), "b" (c)
    : "memory", "ax", "bx"
  );
}

void fputpixel16(char *offset, int c) {
  __asm__ (
    "movw %%bx, (%%eax)"
    :
    : "a" (offset), "b" (c)
    : "memory", "ax", "bx"
  );
}

void fputpixel24(char *offset, int c) {
  __asm__ (
   "andl $0x00FFFFFF,%%ebx;     # mask off the top byte of the colour
    andl $0xFF000000,(%%eax);   # mask off the actual 24-bit value
    orl  %%ebx,(%%eax);         # and or it into place
   "
    :
    : "a" (offset), "b" (c)
    : "memory", "ax", "bx"
  );
}

void fputpixel32(char *offset, int c) {
  __asm__ (
    "movl %%ebx, (%%eax)"
    :
    : "a" (offset), "b" (c)
    : "memory", "ax", "bx"
  );
}

int fgetpixel8(char *offset) {
  int c;
  __asm__ (
    "movb (%%eax), %%bl"
    : "=b" (c)
    : "a" (offset)
    : "memory", "ax", "bx"
  );
  return c & 0x000000FF;
}

int fgetpixel16(char *offset) {
  int c;
  __asm__ (
    "movw (%%eax), %%bx"
    : "=b" (c)
    : "a" (offset)
    : "memory", "ax", "bx"
  );

  return c & 0x0000ffff;
}

int fgetpixel24(char *offset) {
  int c;
  __asm__ (
    "movl (%%eax), %%ebx"
    : "=b" (c)
    : "a" (offset)
    : "memory", "ax", "bx"
  );

  return c & 0x00FFFFFF;
}

int fgetpixel32(char *offset) {
  int c;
  __asm__ (
    "movl (%%eax), %%ebx"
    : "=b" (c)
    : "a" (offset)
    : "memory", "ax", "bx"
  );

  return c;
}
