/* poly.cc
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * Contains different portable polygon drawing functions for PETAL.
 * Because they use hline for their low level drawing, the polygon
 * functions support offsetting and clipping.
 */

#include <iostream.h>
#include "petal.h"
#include "poly.h"

/* our global arrays */
Scanline *scanline;

/* This is basically a modified fixed point line drawing function which will
 * store the line's x's and y's in the global arrays
 */
void rasterize(Point2d *p1, Point2d *p2)	{
  int	x1	= p1->x,  y1 = p1->y;
  int	x2	= p2->x,  y2 = p2->y;
  int	y;
  long x,xinc;

  if(y2	!=	y1) {								/* we don't want a horizontal line      */
	 if(y2 < y1) {			    				/* make sure we're going in the right   */
		y1	^=	y2;								/* direction,  if not swap x,'s and y's */
		y2	^=	y1;
		y1	^=	y2;

		x1	^=	x2;
		x2	^=	x1;
		x1	^=	x2;
	 }

	 x	= (long)x1 << 8;						/* start at our first x */
	 /* this will find what we need to increment to our next x */
	 xinc	= ((long)(x2 -	x1) << 8) / ((long)(y2	- y1));

	 /* Here we skip the first point so that we don't end up storing the
	  * point where two lines meet, twice.  Resulting in a single point drawn.
     */
	 x	+=	xinc;
	 y1++;

	 for(y =	y1; y	<=	y2; y++)	{		             /* loop through each y                  */
		if((y	>=	0)	& (y < petal.display.height))	 /* check if we're in bounds             */
		  if(scanline[y].x1 == 0xFFFF)	          /* if not empty                         */
			 scanline[y].x1 =	x >> 8;				    /* store into the first global array    */
		  else
			 scanline[y].x2 =	x >> 8;					 /* else store in the last global array  */
		x += xinc;								          /* increment to the next x              */
	 }
  }
}

/* Draws a filled polygon given an array of vertices and an polygon structure.
 * Note: There is no need to redclare this for each colour mode!
 */
void poly(Point2d *points, int num_points, int colour, Image *img = petal.default_image) {
  int	i;
  /* Two pointers to a vertex. These are used to connect two vertices
   * together in when calling the polyline routine.
   */
  Point2d *point1, *point2;

  point1	= points;
  point2	= points	+ 1;

  /* Set up our impossible values */
  for	(i	= 0; i <	img->height; i++) scanline[i].x1 = scanline[i].x2 = 0xFFFF;

  for (i = 1; i < num_points; i++) {
    OffsetAdjust(img, point1->x, point1->y);
    /* Calculate the edge of this line */
    rasterize(point1, point2);
    /* then go on to the next line */
    point1 += 1;
    point2 += 1;
  }

  /* now we're nice enough to close the polygon for you */
  point2 = points;
  rasterize(point1, point2);

  for(i = 0; i	< img->height;	i++)			 /* loop through all our y's */
	 if(scanline[i].x1 != 0xFFFF)	{			 /* check for a first and last coordinate */
		if(scanline[i].x2	==	0xFFFF) scanline[i].x2 = scanline[i].x1;				/* just incase only one point was found */
  	   /* Draw a line between the two x coordinates, on the row i, with colour c */
	   hline(scanline[i].x1, scanline[i].x2, i, colour, img);
    }
}
