/* sprite.cc
 * 1997/12/27
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * Contains functions to support a simple sprite system.
 */

#include <fstream.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "petal.h"
#include "sprite.h"
#include "pcx.h"

Sprite::Sprite(int maxx = petal.display.width, int maxy = petal.display.height) {
  bg = new_image(maxx, maxy);
  clear(0, bg);
  num_images = 0;
  screen = new_image(maxx, maxy);
  copy_image(screen, bg);
}

Sprite::~Sprite(void) {
  int i;

  delete_image(bg);
  delete_image(screen);
  for(i = 0; i < num_sprites; i++) delete data[i];
  delete data;
  //delete image;
  delete fps;
  delete current;
  delete status;
  delete sx, sy;
  delete lx, ly;
}

void Sprite::load(char *fname) {
  //ifstream file;
  FILE *file;
  char image_fname[50];
  int i, c;

  cout << "Opening sprite definition: " << fname << endl;
  //file.open(fname);
  file = fopen(fname, "r");
  // check for errors loading the file
  if(file == NULL) {
    cout << "Error opening file\n";
    exit(1);
  }

  // read the in the number of images (what the sprites will be made of)
  //file >> num_images;
  fscanf(file, "%d", &num_images);
  cout << num_images << " images\n";
  image = new Pcx[num_images];
  // Then read each of them in and load the actual PCX
  for(i = 0; i < num_images; i++) {
    //file >> image_fname;
    fscanf(file, "%s", &image_fname);
    cout << "Loading image " << i << ": " << image_fname;
    image[i].load(image_fname);
    cout << " (" << image[i].width << ", " << image[i].height << ")\n";
    transparent(0, image[i].img);
  }

  // read in the number of sprites
  //file >> num_sprites;
  fscanf(file, "%d", &num_sprites);
  cout << num_sprites << " sprites\n";
  // allocate room for many different things:
  //   pointers for each sprites "frame declaration"
  //   number of frames in each sprite
  //   sprite on/off data
  //   x and y data
  //   the last x and y data for dirty rectangles
  //   current sprite frame position
  cout << "Allocating some memory\n";
  data    = new int*[num_sprites];
  fps     = new int [num_sprites];
  status  = new char[num_sprites]; memset(status,0,num_sprites);
  sx      = new int [num_sprites];
  sy      = new int [num_sprites];
  lx      = new int [num_sprites]; memset(lx, 0, num_sprites*4);
  ly      = new int [num_sprites]; memset(ly, 0, num_sprites*4);
  current = new int [num_sprites]; memset(current,0,num_sprites*4);
  // now we'll actually read in the "sprite definition" or "frame declaration"
  cout << "Sprite definition\n";
  for(i = 0; i < num_sprites; i++) {
    //file >> fps[i];
    fscanf(file, "%d", &fps[i]);
    cout << fps[i] << " frames in sprite " << i << endl;
    // allocate room for this sprite definition
    data[i] = new int[ fps[i] ];
    for(c = 0; c < fps[i]; c++) {
      //file >> data[i][c];
      fscanf(file, "%d", &data[i][c]);
      data[i][c]--;      // make it zero based
      cout << data[i][c] << " ";
    }
    cout << endl;
    fps[i]--;  // make it zero based (must be located after use)
  }

  fclose(file);
  return;
}

void Sprite::on(int num) {
  num--;
  if(num < num_sprites) {
    status[num] = 1;
    sx[num] = lx[num] = 0;
    sy[num] = ly[num] = 0;
  }
}

void Sprite::off(int num) {
  num--;
  if(num < num_sprites) {
    status[num] = 0;
    // erase it's leftovers
    /*
    copy_image(lx[num], ly[num],
               image[ data[num][ current[num] ] ].width,
               image[ data[num][ current[num] ] ].height,
               screen,
               bg, lx[num], ly[num]);
    */
    show_image(lx[num], ly[num],
               image[ data[num][ current[num] ] ].width,
               image[ data[num][ current[num] ] ].height,
               bg /*, lx[num], ly[num]*/);
  }
}

void Sprite::set_pos(int num, int x_pos, int y_pos) {
  num--;
  if(num < num_sprites) {
    sx[num] = x_pos;
    sy[num] = y_pos;
  }
}

void Sprite::get_pos(int num, int *x_pos, int *y_pos) {
  num--;
  if(num < num_sprites) {
    *x_pos = sx[num];
    *y_pos = sy[num];
  }
}

void Sprite::update(int tx = 0, int ty = 0) {
  int i, x, y, xadd, yadd;

  // Erase all the sprites first.  Previously I erased sprites, and then drew them
  // directly after in a loop.  This was fine, until sprites overlapped.  What
  // would happen is a transparent sprite would show up over another sprite, but
  // instead of seeing the second sprite through the transparency, you saw the
  // background!  Obviously that wasn't good enough, so I seperated the loops
  for(i = 0; i < num_sprites; i++)
    if(status[i]) {
      // Erase the old sprite by copying a background chunk to the screen buffer
      //cout << "Erase\n";
      copy_image(lx[i], ly[i],
                 image[ data[i][ current[i] ] ].width,
                 image[ data[i][ current[i] ] ].height,
                 screen,
                 bg /*, lx[i], ly[i] */ );
    }
  
  for(i = 0; i < num_sprites; i++) {
    if(status[i]) {
      // This is where I previously erased a sprite

      // Display the new sprite in it's new location
      //cout << "Copy\n";
      copy_image(screen, image[ data[i][current[i]] ].img, sx[i], sy[i]);

      // Now show the whole image.  This is a bit odd, but very easy to understand.
      // What I do is copy the area occupied by the current sprite and where the
      // old one was.  Why?  Because then it only takes one pass (less flicker,
      // more speed) to erase the old sprite, and draw the new one.
      x = MIN(lx[i], sx[i]);
      y = MIN(ly[i], sy[i]);
      xadd = ABS( (sx[i] - lx[i]) ) + 1;
      yadd = ABS( (sy[i] - ly[i]) ) + 1;
      //cout << "Show\n";
      show_image(x, y,
                 image[ data[i][ current[i] ] ].width  + xadd,
                 image[ data[i][ current[i] ] ].height + yadd,
                 screen,
                 /*x + */tx, /* y + */ ty);

      // goto the next frame of the sprite
      if(current[i] >= fps[i]) current[i] = 0;
      else current[i]++;

      // Update the sprite's old location
      lx[i] = sx[i];
      ly[i] = sy[i];
    }
  }
}
